/* This file contains routine to find a warping described by xmorph like
   meshes. Source mesh points are mapped to destination mesh points.
   Both meshes should be topologically equivalent to Cartesian grids. */

/* This code is taken from xmorph sources and slightly modified. */
/* Jan Kybic, November 1998 */

/* See George Wolberg's "Digital Image Warping"
   IEEE Computer Society Press order number 1944 */

#ifndef MAKINGMEX
#include <stdio.h>
#define MY_CALLOC(s,t) (void *)malloc(s*sizeof(t))
#define FREE(x) mxFree(x)
#endif




#define MIN(a,b) ((a)<(b)?(a):(b))
#define ABS(a)   ((a)>=0?(a):(-(a)))
#define MAX(x,y) ((x)>(y) ? (x) : (y))
#define NEAR(x1, x2) (((x2)!=0.0) && (((x1)/(x2)) >= 0.999) && (((x1)/(x2))<1.001))

#define REAL    double
#ifndef FALSE
#define FALSE 0
#endif


/* --------------------------------------------------------------- */
/* NAME
//   d_parabola -- returns the derivative of a parabola fit through 3 points
//
// To find this formula:
//   A parabola is
//     y = a x^2 + b x + c.  (eqn 1)
//   Write this for 3 points, (x0,y0) (x1,y1), (x2,y2):
//     y0 = a x0^2 + b x0 + c  (eqn 2 i)
//     y1 = a x1^2 + b x1 + c  (eqn 2 ii)
//     y2 = a x2^2 + b x2 + c  (eqn 2 iii)
//   Solve this system for the constants a, b, c.
//   Substitute these a, b, c back into (eqn 1).  Take the derivative of
//   (eqn 1).
*/

REAL
d_parabola(REAL x, REAL xp0, REAL xp1, REAL xp2,
           REAL yp0, REAL yp1, REAL yp2)
{
  REAL dP=(  xp0*(yp1-yp2)*(2.0*x - xp0)
             + xp1*(yp2-yp0)*(2.0*x - xp1)
             + xp2*(yp0-yp1)*(2.0*x - xp2)) / ((xp0-xp1)*(xp0-xp2)*(xp2-xp1));

  return dP;
}




/* --------------------------------------------------------------- */
/* NAME
//   hermite3_interp : cubic hermite interpolation
//
//
// ARGUMENTS
//   w (in): evaluation abcissa
//     w should be in the range x[0] <= w <= x[n-1]
//
//   x (in): abcissas at knots
//     abcissas should be ordered such that x[0] < x[1] < ... < x[n-1]
//
//   y (in): ordinates at knots
//
//   d (in/out): derivatives at knots
//
//   n: (in): number of knots
//
//   f: (in): derivative function (or NULL if not available)
//
//   flags (in): bit field:
//     (flags&1): compute_deriv:
//                0 => use d as input, 1 => compute d
//                NOTE: if compute_deriv == 1 and f == NULL then the
//                      derivatives are estimated using a parabola fit.
//     (flags&2): periodic:
//                0 => non-periodic domain, 1 => periodic domain
//                NOTE:  for periodic domain, the abcissa of the
//                  point outside of the explicitly provided domain
//                  is estimated by using the distance between the
//                  outer 2 points to find the distance to the next
//                  point.
//   h1 (out): first derivative of spline, NULL=>ignore
//
//   h2 (out): second derivative of spline, NULL=>ignore
//
//
// DESCRIPTION
//   This interpolant is cubic (i.e. is a third order polynomial) fit
//   through two knots where the values of the function and the values of
//   the first derivative of the function are known at the knots.
//   (Stricly speaking, the values of the derivatives do not need to be
//   analytically known, since some estimate of the derivative can be
//   used.)
//
//
// RETURN VALUE
//   Evaluation of interpolation
//
//
// NOTES
//   The estimated derivatives can be pretty lousy.
//
//   To derive this formula:
//     Call the interpolant H(x).  Name the values
//       H(x_i)     = y_i        H'(x_i)     = d_i
//       H(x_(i+1)) = y_(i+1)    H'(x_(i+1)) = d_(i+1)
//     The interpolant has the form
//       H(x) = y_i + h_i d_i + h_i^2 A + h_i^2 B (x - x_(i+1))
//     where
//       h_i = x_(i+1) - x_i
//     The derivative therefore has the form
//       H'(x) = d_i + 2 A h_i + B (2 h_i (x - x_(i+1)) + h_i^2)
//     Now substitute in the values at x_(i+1) into H and H', and set them
//     equal to y_(i+1) and d_(i+1) respectively to get A and B.
//
//     The interpolant is a proper function, where x is the abcissa and
//     y is the ordinate.  However, if data is provided such that
//     the interpolant is not a function such that each abcissa
//     has exactly one ordinate (i.e. where some subset of x values
//     overlap), then the interpolant will give incorrect results.
//     If this is the case then the data you are trying to interpolate
//     can not be represented by a proper function, so you will have
//     to parameterize the data with some third parameter variable
//     (usually refered to as t), and interpolate x and y with respect
//     to that third variable.
*/
REAL
hermite3_interp(REAL w, const REAL *x, const REAL *y, REAL *d, long n,
                REAL (*f)(REAL), int flags, REAL *h1, REAL *h2)
{
  double A, B;
  double h, h_2;
  double diff=0.0;
  double H;
  long si;
  int compute_deriv = (flags&1);
  int periodic = (flags&2);

#if (VERBOSE >= 3)
  printf("hermite3_interp: at %12g\n", w);
  for(si=0; si < n; si++) {
    printf("hermite3_interp: [%4li]  x = % 12g  y = % 12g\n", si, x[si], y[si]);
  }
#endif

  /* Find interval of spline to evaluate
  // I.e. find the index such that the evaluation abcissa, w, is to
  // the right of the abcissa associated with that index.  Then, x is
  // between x[si] and x[si+1].
  */
  /* MJG 14jul95 -- do not start at si=n-1.  See "h =" line below.
  // MJG 18jul94 -- was reading beyond bounds at last knot. (maybe)
  */
  for(si=n-2; (si>=0) && ((diff=(w-x[si])) < 0.0); si--)
    ;

  /* We are sitting at a knot so no interpolation needed */
  if(0.0 == diff) {
#if (VERBOSE >= 3)
    printf("hermite3_interp: at knot %li\n", si);
#endif
    return y[si];
  }

  /* h is the interval between knots */
  h   = x[si+1] - x[si];
  h_2 = h*h;

#if (VERBOSE >= 3)
  printf("hermite3_interp: si = %li    diff = %12g\n", si, diff);
  printf("hermite3_interp: h = %12g    h^2 = %12g\n", h, h_2);
#endif

  /* either the derivatives were provided or must be found */
  if(compute_deriv) {
    /* must calculate derivatives */
    if(f != NULL) {
      /* calculate the derivative */
      d[si]=(*f)(x[si]);
      d[si+1]=(*f)(x[si+1]);

    } else {
      /* approximate derivative using parabola fit */
      if(0 == si) {
        /* at first knot */
        d[si+1] = d_parabola(x[si+1], x[si], x[si+1], x[si+2],
                                      y[si], y[si+1], y[si+2]);
        if(!periodic) {
          d[si] = d_parabola(x[si], x[si], x[si+1], x[si+2],
                                    y[si], y[si+1], y[si+2]);
        } else { /* periodic */
          /* estimate left-end abcissa */
          REAL xn = 2.0 * x[si] - x[si+1];
          d[si] = d_parabola(x[si], xn,     x[si], x[si+1],
                                    y[n-1], y[si], y[si+1]);
        }

      } else if(si >= (n-2)) {
        /* at last or 2nd to last knot */
        d[si] = d_parabola(x[si], x[si-1], x[si], x[si+1],
                                  y[si-1], y[si], y[si+1]);
        if(!periodic) {
          d[si+1] = d_parabola(x[si+1], x[si-1], x[si], x[si+1],
                                        y[si-1], y[si], y[si+1]);
        } else { /* periodic */
          /* estimate right-end abcissa */
          REAL xn = 2.0 * x[si+1] - x[si];
          d[si+1] = d_parabola(x[si+1], x[si], x[si+1], xn,
                                        y[si], y[si+1], y[0]);
        }

      } else {
        /* between first and 2nd to last knot */
        d[si] = d_parabola(x[si], x[si-1], x[si], x[si+1],
                                  y[si-1], y[si], y[si+1]);
        d[si+1] = d_parabola(x[si+1], x[si], x[si+1], x[si+2],
                                      y[si], y[si+1], y[si+2]);
      }
    }
  }

  /* calculate interpolant parameters */
  A = (y[si+1] - y[si] - h*d[si])/h_2;
  B = (d[si+1] - d[si] - 2.0*h*A)/h_2;

  /* evaluate spline derivatives */
  if(h1!=NULL) *h1 = d[si] + diff*(2.0*A + B*(diff + 2.0*(w-x[si+1])));
  if(h2!=NULL) *h2 = 2.0*A + 2.0*B*(2.0*diff + (w-x[si+1]));

  /* return the spline evaluation */
  H  =  y[si] + diff*(d[si] + diff*(A + (w-x[si+1])*B));

  return H;
}

/* --------------------------------------------------------------- */

/* derivative_hack: compute 1st derivative of x,y data (len entries)
//
// Written and Copyright (C) 1994-1997 by Michael J. Gourlay
//
// PROVIDED AS IS.  NO WARRANTEES, EXPRESS OR IMPLIED.
//
// Mathematically, it's a hack to prevent overshooting knots, but
// maintain smoothness and it works more intuitively.  Besides, we all
// know that mathematicians are too worried about rigor, and the
// physicists end up creating the right math. -- MJG
*/
static void derivative_hack(const double *x, const double *y, double *yd, int len)
{
  int indx;

  if(x[0]==x[1]) {
    yd[0] = 0.0; /* avoid division by zero */
  } else {
    yd[0] = (y[1]-y[0])/(x[1]-x[0]);
  }
  if(x[len-2] == x[len-1]) {
    yd[len-1] = 0.0; /* avoid division by zero */
  } else {
    yd[len-1] = (y[len-1]-y[len-2])/(x[len-1]-x[len-2]);
  }
  for(indx=1; indx<(len-1); indx++) {
    if(x[indx-1]==x[indx] || x[indx]==x[indx+1]) {
      yd[indx] = 0.0; /* avoid division by zero */
    } else {
      if( (y[indx-1] - y[indx]) * (y[indx+1] - y[indx]) >= 0.0) {
        /* There was a change in the sign of yd so force it zero */
        /* This will prevent the spline from overshooting this knot */
        yd[indx] = 0.0;
      } else {
        /* Set slope at this knot to slope between two adjacent knots */
        yd[indx] = (y[indx-1]-y[indx+1]) / (x[indx-1]-x[indx+1]);
      }
    }
  }
}




/* --------------------------------------------------------------- */

/* hermite3_array : cubic hermite interpolation for an array of points
//
// Uses derivative_hack to find derivatives.
//
//
// kx, ky (in):  arrays of knots
// nk (in):      number of knots
// sx (in):      evaluation points array
// sy (out):     spline values at evaluation pts sx.  Must already be allocated.
// ns (in):      number of evaluation points
*/
int
hermite3_array(const double *kx, const double *ky, long nk, double *sx,
double *sy, long ns)
{
  register long xi;
  double *kyd;

#if 0
  mexPrintf("hermite3_array called with:\n") ;
  for(xi=0;xi<nk;xi++) mexPrintf("(%g,%g) ",kx[xi],ky[xi]) ;
  mexPrintf("\n") ;
#endif 

  if((kyd=MY_CALLOC(nk, double))==NULL)
    return 1;

  /* Test bounds. */
  /* As of 18jul94, this test was triggering for cases
  // where the bounds were nearly equal, but slightly out of range, in
  // which case the spline should work anyway, which is why I let it run
  // even if the spline abcissas are out of range.
  */
  if((sx[0] < kx[0]) || (sx[ns-1] > kx[nk-1])) {
    if(!NEAR(sx[ns-1], kx[nk-1])) {
#if 1
      printf("hermite3_array: out of range:\n");
      printf("hermite3_array: eval=%.20g < knot=%.20g | %.20g>%.20g\n",
        sx[0], kx[0], sx[ns-1], kx[nk-1]);
#endif 
    }
  }

  /* Find array of derivatives */
  derivative_hack(kx, ky, kyd, nk);

  /* Evaluate the spline */
  for(xi=0; xi<ns; xi++) {
    sy[xi]=hermite3_interp(sx[xi], (const REAL *)kx, (const REAL *)ky, 
			   (REAL *)kyd, nk, NULL, FALSE, NULL, NULL);
  }

  /* Free the derivatives array */
  FREE(kyd);

#if 0
  mexPrintf("hermite3_array returned with:\n") ;
  for(xi=0;xi<ns;xi++) mexPrintf("%g(%g,%g) ",sx[xi],sy[xi]) ;
  mexPrintf("\n") ;
#endif

  return 0;
}



/* --------------------------------------------------------------- */

/* NAME
     xmorphwarp: 2-pass mesh-based single-channel image warping
  
  
   ARGUMENTS
     img_width (in)  : width dimension, in pixels, of in and out images
     img_height (in) : height dimension, in pixels, of in and out images
     xs (in)         : source mesh x-coordinate values
     ys (in)         : source mesh y-coordinate values
     xd (in)         : destination mesh x-coordinate values
     yd (in)         : destination mesh y-coordinate values
     mesh_width (in) : x-dimension of meshes
     mesh_height (in): y-dimension of meshes
     warpx (out)     : x component of the warp for each pixel
     warpy (out)     : y component of the warp for each pixel
  
  The output array must be allocated (img_height*img_width*sizeof(double) bytes 
  each) before this routine is called. 

   NOTES
     Mesh points must not fold over,
     and border mesh points must stay on border
  
   AUTHOR
     This code was originally written by George Wolberg, based on Smythe90.
     Modifications and many bug fixes by Michael J. Gourlay and other
     authors as noted in the code comments.
     Modified by Jan Kybic  
  
*/
void
xmorphwarp(int img_width, int img_height,
  const double *xs, const double *ys, const double *xd, const double *yd,
  int mesh_width, int mesh_height,double *warpx, double *warpy)
{
  /* "const type *" means "pointer to a constant array of type"
     not "constant pointer to type"
  */

  int          ai;
  int          xi, yi;
  const double *x1, *y1, *x2, *y2;
  double       *xrow1, *yrow1, *xrow2, *yrow2;
  double       *map1, *map2;
  double       *indx;
  double       *ts, *ti, *td;


  /* allocate memory for buffers:
    
     indx stores indices used to sample splines
     xrow1, xrow2, yrow2, yrow2 store column data in row order for spline
     map1, map2 store mapping functions computed in row order in spline
    
     Could use alloca instead to avoid the free's at the end of this
     routine.  alloca might also be faster.  In fact, the original code
     might have used alloca.  I do not remember why I changed it.
  */

  ai    = MAX(img_height, img_width) + 1;
  indx  =  MY_CALLOC(ai, double);
  xrow1 =  MY_CALLOC(ai, double);
  yrow1 =  MY_CALLOC(ai, double);
  xrow2 =  MY_CALLOC(ai, double);
  yrow2 =  MY_CALLOC(ai, double);
  map1  =  MY_CALLOC(ai, double);
  map2  =  MY_CALLOC(ai, double);


  /* First pass (phase one): create tables ts and ti for x-intercepts
     of vertical splines in S and I.
     tables have mesh_width columns of height img_height
  */
  ts = MY_CALLOC(mesh_width*img_height, double);
  ti = MY_CALLOC(mesh_width*img_height, double);
  for(yi=0; yi<img_height; yi++) /* indices used to sample vert spline */
    indx[yi] = yi;

  for(xi=0; xi<mesh_width; xi++) { /* visit each vertical spline */
    /* store columns as rows for spline */
    for(yi=0; yi<mesh_height; yi++) {
      xrow1[yi] = xs[yi*mesh_width+xi];
      xrow2[yi] = xd[yi*mesh_width+xi];
      yrow1[yi] = ys[yi*mesh_width+xi];
      yrow2[yi] = yd[yi*mesh_width+xi];
    }

    /* scan convert vertical splines of S and I */

    hermite3_array(yrow1, xrow1, mesh_height, indx, map1, img_height);
    hermite3_array(yrow1, xrow2, mesh_height, indx, map2, img_height);

    /* store resampled rows back into columns */
    for(yi=0; yi<img_height; yi++) {
      ts[yi*mesh_width+xi] = map1[yi];
      ti[yi*mesh_width+xi] = map2[yi];
    }
  }

  /* First pass (phase two): warp x using ts and ti.
     tmp holds intermediate image.
  */
  for(xi=0; xi < img_width; xi++) /* indices used to sample horiz spline */
    indx[xi]=xi;

  for(yi=0; yi < img_height; yi++) { /* visit each row */
    /* fit spline to x-intercepts; resample over all columns */
    x1 = &ts[yi*mesh_width];
    x2 = &ti[yi*mesh_width];
    hermite3_array(x1, x2, mesh_width, indx, warpx+yi*img_width, img_width);
  }
  /* free buffers */
  FREE(ts);
  FREE(ti);

  /* Second pass (phase one): create tables ti and td for y-intercepts
     of horiz splines in I and D.
     Tables have mesh_height rows of width img_width
  */
  ti = MY_CALLOC(mesh_height*img_width, double);
  td = MY_CALLOC(mesh_height*img_width, double);

  for(xi=0; xi < img_width; xi++)
    indx[xi] = xi; /* indices used to sample horiz splines */

  for(yi=0; yi < mesh_height; yi++) {

    /* scan convert horizontal splines of I and D */
    y1 = &ys[yi*mesh_width];
    x2 = &xd[yi*mesh_width];
    y2 = &yd[yi*mesh_width];

    hermite3_array(x2, y1, mesh_width, indx, &ti[yi*img_width], img_width);
    hermite3_array(x2, y2, mesh_width, indx, &td[yi*img_width], img_width);
  }

  /* Second pass (phase two): warp y using ti and td */
  for(yi=0; yi < img_height; yi++) indx[yi] = yi;
  for(xi=0; xi < img_width; xi++) {
    /* store columns as row for hermite3_array */
    for(yi=0; yi<mesh_height; yi++) {
      xrow1[yi] = ti[yi*img_width+xi];
      yrow1[yi] = td[yi*img_width+xi];
    }

    /* fit spline to y-intercepts:  resample over all rows */
    hermite3_array(xrow1, yrow1, mesh_height, indx, map1, img_height);

    /* transpose again */
    for(yi=0; yi<img_height; yi++) warpy[xi+yi*img_width]=map1[yi] ;
  }

  FREE(ti);
  FREE(td);
  FREE(indx);
  FREE(xrow1);
  FREE(yrow1);
  FREE(xrow2);
  FREE(yrow2);
  FREE(map1);
  FREE(map2);
}
