/* The following functions correspond to their Matlab equivalents */

double cbspln(double x)
{ 
  x=fabs(x) ;
  if (x>2.0) return 0 ;
  if (x>1) return (2-x)*(2-x)*(2-x)/6.0 ;
  return 2.0/3.0-x*x*(1-0.5*x) ;
}

double qbspln(double x)
{ 
  x=fabs(x) ;
  if (x>1.5) return 0 ;
  if (x>0.5) return (1.5-x)*(1.5-x)/2.0 ;
  return 0.75-x*x ;
}

double qbderiv(double x)
{ double xa ;
  xa=fabs(x) ;
  if (xa>1.5) return 0 ;
  if (xa>0.5) return (x>0) ? xa-1.5 : 1.5-xa ;
  return -2*x ;
}

double qbdderiv(double x)
{ 
  if (x>0) {
    if (x>1.5) return 0 ;
    if (x>0.5) return 1 ; else return -2 ;
  } else {
    if (x<-1.5) return 0 ;
    if (x<-0.5) return 1 ; else return -2 ;
  } 
}



double lbspln(double x)
{ 
  x=fabs(x) ;
  if (x>1) return 0 ;
  return 1-x ;
}

double lbderiv(double x)
{ 
  if (fabs(x)>1) return 0 ;
  if (x>0) return -1 ; else return 1 ;
}
  
double cbderiv(double x)
{ double xa,y ;

  y=x>0 ? 1 : -1 ;
  x=fabs(x) ;
  if (x>2) return 0 ;
  if (x>1) return -0.5*y*(2-x)*(2-x) ;
  return (1.5*x-2)*x*y ;
}

double cbdderiv(double x)
{ 
  x=fabs(x) ;
  if (x>2) return 0.0 ;
  if (x>1) return 2.0-x ;
  return 3.0*x-2.0 ;
}

double cbinterp(double *c,int nc,double x)
/* Given cubic B-spline coefficients cp (nc of them) at 1,2,..,nc
   find value at point x */
{ int xf ; double y ;
  xf=(int)floor(x)-2 ;
  
  if (xf>=0 && xf<nc) y=c[xf]*cbspln(x-xf-1) ;
  else y=0.0 ;
  if (xf+1>=0 && xf+1<nc) y+=c[xf+1]*cbspln(x-xf-2) ;  
  if (xf+2>=0 && xf+2<nc) y+=c[xf+2]*cbspln(x-xf-3) ;  
  if (xf+3>=0 && xf+3<nc) y+=c[xf+3]*cbspln(x-xf-4) ;  
  return y ;
}

double qbinterp(double *c,int nc,double x)
/* Given quadratic B-spline coefficients c (nc of them) at 1,2,..,nc
   find value at point x */
{ int xf ; double y ;
  xf=(int)floor(x-1.5) ;
  
  if (xf>=0 && xf<nc) y=c[xf]*qbspln(x-xf-1) ;
  else y=0.0 ;
  if (xf+1>=0 && xf+1<nc) y+=c[xf+1]*qbspln(x-xf-2) ;  
  if (xf+2>=0 && xf+2<nc) y+=c[xf+2]*qbspln(x-xf-3) ;  
  return y ;
}

double lbinterp(double *c,int nc,double x)
/* Given linear B-spline coefficients c (nc of them) at 1,2,..,nc
   find value at point x */
{ int xf ; double y ;
  xf=(int)floor(x-1) ;
  
  if (xf>=0 && xf<nc) y=c[xf]*lbspln(x-xf-1) ;
  else y=0.0 ;
  if (xf+1>=0 && xf+1<nc) y+=c[xf+1]*lbspln(x-xf-2) ;  
  return y ;
}

double lbinterpd(double *c,int nc,int nd,double x,double y)
/* Given linear B-spline coefficients c (nc columns and nd rows
stored columnwise at (1,2,...,nc)x(1,2,...,nd) grid, find value at point 
x,y */
{ int xf ; double r ;
  xf=(int)floor(x) ;
  
  if (xf>=1 && xf<=nc) r=lbinterp(c+(xf-1)*nd,nd,y)*lbspln(x-xf) ;
  else r=0.0 ;
  if (xf+1>=1 && xf+1<=nc) r+=lbinterp(c+(xf)*nd,nd,y)*lbspln(x-xf-1) ;  
  return r ;
}

double qbinterpd(double *c,int nc,int nd,double x,double y)
/* Given quadratic B-spline coefficients c (nc columns and nd rows
stored columnwise at (1,2,...,nc)x(1,2,...,nd) grid, find value at point 
x,y */
{ int xf ; double r ;
  xf=(int)floor(x-0.5) ;
  
  if (xf>=1 && xf<=nc) r=qbinterp(c+(xf-1)*nd,nd,y)*qbspln(x-xf) ;
  else r=0.0 ;
  if (xf+1>=1 && xf+1<=nc) r+=qbinterp(c+(xf)*nd,nd,y)*qbspln(x-xf-1) ;  
  if (xf+2>=1 && xf+2<=nc) r+=qbinterp(c+(xf+1)*nd,nd,y)*qbspln(x-xf-2) ;  
  return r ;
}

double cbinterpd(double *c,int nc,int nd,double x,double y)
/* Given cubic B-spline coefficients c (nc columns and nd rows
stored columnwise at (1,2,...,nc)x(1,2,...,nd) grid, find value at point 
x,y */
{ int xf ; double r ;
  xf=(int)floor(x-1.0) ;
  
  if (xf>=1 && xf<=nc) r=cbinterp(c+(xf-1)*nd,nd,y)*cbspln(x-xf) ;
  else r=0.0 ;
  if (xf+1>=1 && xf+1<=nc) r+=cbinterp(c+(xf)*nd,nd,y)*cbspln(x-xf-1) ;  
  if (xf+2>=1 && xf+2<=nc) r+=cbinterp(c+(xf+1)*nd,nd,y)*cbspln(x-xf-2) ;  
  if (xf+3>=1 && xf+3<=nc) r+=cbinterp(c+(xf+2)*nd,nd,y)*cbspln(x-xf-3) ;  
  return r ;
}


void imwarpbtd(double *yp,double *fp,double *cp,int n,int nf,int nc,int ny) 
/* warp an image given by cubic B-spline interpolation coeficients f
  of nf columns and ny rows (stored in Matlab/Fortran fashion) using
  nc quadratic B-spline coefficients cp. Create resulting image yp of
  n columns. */
{ int i,j ; double r,*tf ;

 /* transpose fp */
 SAFEMALLOC(tf,ny,nf) ;
 for(i=0;i<nf;i++) /* iterate over columns */
   for(j=0;j<ny;j++) /* iterate over rows */
     tf[j*nf+i]=fp[j+ny*i] ;

 for(i=0;i<n;i++) { /* iterate over columns */
   r=2+((double)i/(n-1))*(nc-3) ;
   r=qbinterp(cp,nc,r) ;
   for (j=0;j<ny;j++) /* iterate over rows */
     yp[j+ny*i]=cbinterp(tf+j*nf,nf,r) ;
 }
 mxFree(tf) ;  
}


#if 1
/* Here you can choose among two equivalent implementations of imwarpdd,
   the first one should be faster. */


void imwarpbdd(double *yp,double *fp,double *cp,int n,int nf,int nc,int nd,
	       int ny) 
/* warp an image given by cubic B-spline interpolation coeficients f
  of nf columns and ny rows (stored in Matlab/Fortran fashion) using
  nc times nd quadratic B-spline coefficients cp (stored
  columnwise). Create resulting image yp of n columns. */
{ int i,j ; double r,*tf,*row ;

 SAFEMALLOC(tf,ny,nf) ;
 SAFEMALLOC(row,ny,1) ;

 /* transpose fp */
 for(i=0;i<nf;i++) /* iterate over columns */
   for(j=0;j<ny;j++) /* iterate over rows */
     tf[j*nf+i]=fp[j+ny*i] ;

 for(j=0;j<ny;j++) { /* iterate over rows */
   r=2+((double)j/(ny-1))*(nd-3) ;
   for(i=0;i<nc;i++) row[i]=qbinterp(cp+i*nd,nd,r) ;
   for (i=0;i<n;i++) { /* iterate over columns */
     r=2+((double)i/(n-1))*(nc-3) ;
     r=qbinterp(row,nc,r) ;
     yp[j+ny*i]=cbinterp(tf+j*nf,nf,r) ;
   }
 }
 mxFree(tf) ;  mxFree(row) ;
}
#else

void imwarpbdd(double *yp,double *fp,double *cp,int n,int nf,int nc,int nd,
	       int ny) 
/* warp an image given by cubic B-spline interpolation coeficients f
  of nf columns and ny rows (stored in Matlab/Fortran fashion) using
  nc times nd quadratic B-spline coefficients cp (stored
  columnwise). Create resulting image yp of n columns. */
{ int i,j ; double r,rx,ry,*tf ;

 SAFEMALLOC(tf,ny,nf) ;

 /* transpose fp */
 for(i=0;i<nf;i++) /* iterate over columns */
   for(j=0;j<ny;j++) /* iterate over rows */
     tf[j*nf+i]=fp[j+ny*i] ;

 for(j=0;j<ny;j++) { /* iterate over rows */
   ry=2+((double)j/(ny-1))*(nd-3) ;
   for (i=0;i<n;i++) { /* iterate over columns */
     rx=2+((double)i/(n-1))*(nc-3) ;
     r=qbinterpd(cp,nc,nd,rx,ry) ;
     yp[j+ny*i]=cbinterp(tf+j*nf,nf,r) ;
   }
 }
 mxFree(tf) ;  
}

#endif

void imwarpbddh(double *yp,double *fp,double *cp,int n,int nf,int nc,int nd,
	       int ny) 
/* warp an image given by cubic B-spline interpolation coeficients f
  of nf columns and ny rows (stored in Matlab/Fortran fashion) using
  nc times nd quadratic B-spline coefficients cp (stored
  columnwise). Create resulting image yp of n columns. 
  Like imwarpbdd, only the warp functions are shifted by 0.5 instead of 1 */
{ int i,j ; double r,rx,ry,*tf ;

 SAFEMALLOC(tf,ny,nf) ;

 /* transpose fp */
 for(i=0;i<nf;i++) /* iterate over columns */
   for(j=0;j<ny;j++) /* iterate over rows */
     tf[j*nf+i]=fp[j+ny*i] ;

 for(j=0;j<ny;j++) { /* iterate over rows */
   ry=1.5+((double)j/(ny-1))*(nd-2) ;
   for (i=0;i<n;i++) { /* iterate over columns */
     rx=1.5+((double)i/(n-1))*(nc-2) ;
     r=qbinterpd(cp,nc,nd,rx,ry) ;
     yp[j+ny*i]=cbinterp(tf+j*nf,nf,r) ;
   }
 }
 mxFree(tf) ;  
}

void imwarpldd(double *yp,double *fp,double *cp,int n,int nf,int nc,int nd,
	       int ny) 
/* warp an image given by cubic B-spline interpolation coeficients f
  of nf columns and ny rows (stored in Matlab/Fortran fashion) using
  nc times nd linear B-spline coefficients cp (stored
  columnwise). Create resulting image yp of n columns. 
*/
{ int i,j ; double r,rx,ry,*tf ;

 SAFEMALLOC(tf,ny,nf) ;

 /* transpose fp */
 for(i=0;i<nf;i++) /* iterate over columns */
   for(j=0;j<ny;j++) /* iterate over rows */
     tf[j*nf+i]=fp[j+ny*i] ;

 for(j=0;j<ny;j++) { /* iterate over rows */
   ry=1+((double)j/(ny-1))*(nd-1) ;
   for (i=0;i<n;i++) { /* iterate over columns */
     rx=1+((double)i/(n-1))*(nc-1) ;
     r=lbinterpd(cp,nc,nd,rx,ry) ;
     yp[j+ny*i]=cbinterp(tf+j*nf,nf,r) ;
   }
 }
 mxFree(tf) ;  
}

void imwarpcdd(double *yp,double *fp,double *cp,int n,int nf,int nc,int nd,
	       int ny) 
/* warp an image given by cubic B-spline interpolation coeficients f
  of nf columns and ny rows (stored in Matlab/Fortran fashion) using
  nc times nd cubic B-spline coefficients cp (stored
  columnwise). Create resulting image yp of n columns. */
{ int i,j ; double r,*tf,*row ;

 SAFEMALLOC(tf,ny,nf) ;
 SAFEMALLOC(row,ny,1) ;

 /* transpose fp */
 for(i=0;i<nf;i++) /* iterate over columns */
   for(j=0;j<ny;j++) /* iterate over rows */
     tf[j*nf+i]=fp[j+ny*i] ;

 for(j=0;j<ny;j++) { /* iterate over rows */
   r=2.0+((double)j/(ny-1))*(nd-3) ;
   for(i=0;i<nc;i++) row[i]=cbinterp(cp+i*nd,nd,r) ;
   for (i=0;i<n;i++) { /* iterate over columns */
     r=2.0+((double)i/(n-1))*(nc-3) ;
     r=cbinterp(row,nc,r) ;
     yp[j+ny*i]=cbinterp(tf+j*nf,nf,r) ;
   }
 }
 mxFree(tf) ;  mxFree(row) ;
}


void imwarpdcdd(double *yp,double *fp,double *cp,int nfx,int nfy,
		int ntx,int nty,int nc,int nd) 
/* warp an image given by cubic B-spline interpolation coeficients f
  of nfx columns and nfy rows (stored in Matlab/Fortran fashion) using
  nc times nd times 2 cubic B-spline coefficients cp (stored
  columnwise). Create resulting image yp of ntx columns and nty rows. */
{ int i,j ; double r,rx,ry,*rowx,*rowy ;

 SAFEMALLOC(rowx,nc,1) ;
 SAFEMALLOC(rowy,nc,1) ;

 for(j=0;j<nty;j++) { /* iterate over rows */
   r=2.0+((double)j/(nty-1))*(nd-3) ; 
   for(i=0;i<nc;i++) rowx[i]=cbinterp(cp+i*nd,nd,r) ;
   for(i=0;i<nc;i++) rowy[i]=cbinterp(cp+nc*nd+i*nd,nd,r) ;

   for (i=0;i<ntx;i++) { /* iterate over columns */
     r=2.0+((double)i/(ntx-1))*(nc-3) ;
     rx=cbinterp(rowx,nc,r) ;
     ry=cbinterp(rowy,nc,r) ;
     yp[j+nty*i]=cbinterpd(fp,nfx,nfy,rx,ry) ;
   }
 }
  mxFree(rowx) ; mxFree(rowy) ;
}


void imwarpcdddeg(double *yp,double *fp,double *cp,int n,int nf,int nc,int nd,
	       int ny,int deg) 
/* warp an image given by deg-degree B-spline interpolation coeficients f
  of nf columns and ny rows (stored in Matlab/Fortran fashion) using
  nc times nd cubic B-spline coefficients cp (stored
  columnwise). Create resulting image yp of n columns. */
{ int i,j ; double r,*tf,*row ;

 SAFEMALLOC(tf,ny,nf) ;
 SAFEMALLOC(row,ny,1) ;

 /* transpose fp */
 for(i=0;i<nf;i++) /* iterate over columns */
   for(j=0;j<ny;j++) /* iterate over rows */
     tf[j*nf+i]=fp[j+ny*i] ;

 switch (deg) {
 case 3:
   for(j=0;j<ny;j++) { /* iterate over rows */
     r=2.0+((double)j/(ny-1))*(nd-3) ;
     for(i=0;i<nc;i++) row[i]=cbinterp(cp+i*nd,nd,r) ;
     for (i=0;i<n;i++) { /* iterate over columns */
       r=2.0+((double)i/(n-1))*(nc-3) ;
       r=cbinterp(row,nc,r) ;
       yp[j+ny*i]=cbinterp(tf+j*nf,nf,r) ;
     }
   }
   break ;
 case 2:
   for(j=0;j<ny;j++) { /* iterate over rows */
     r=2.0+((double)j/(ny-1))*(nd-3) ;
     for(i=0;i<nc;i++) row[i]=cbinterp(cp+i*nd,nd,r) ;
     for (i=0;i<n;i++) { /* iterate over columns */
       r=2.0+((double)i/(n-1))*(nc-3) ;
       r=cbinterp(row,nc,r) ;
       yp[j+ny*i]=qbinterp(tf+j*nf,nf,r) ;
     }
   }
   break ;
 case 1:
   for(j=0;j<ny;j++) { /* iterate over rows */
     r=2.0+((double)j/(ny-1))*(nd-3) ;
     for(i=0;i<nc;i++) row[i]=cbinterp(cp+i*nd,nd,r) ;
     for (i=0;i<n;i++) { /* iterate over columns */
       r=2.0+((double)i/(n-1))*(nc-3) ;
       r=cbinterp(row,nc,r) ;
       yp[j+ny*i]=lbinterp(tf+j*nf,nf,r) ;
     }
   }
   break ;
 }
 mxFree(tf) ;  mxFree(row) ;
}

