function x=conjgradmin(critfun,x0)
% Conjugated gradient minimization method
% critfun - criterion function, e.g. 'crit', will be called
% with one argument - a vector x. Should return two output argument: 
% the criterion and
% its gradient with respect to x.
% x0 is the starting guess
%% Jan Kybic, 1999

x=x0(:) ;
maxiter=500 ; % maximum number of outer iterations
miniter=2 ; % minimum number of outer iterations
prec=1e-4 ; % outer loop precision

for i=1:maxiter,
[J0,gr]=feval(critfun,x) ; % J0 
gr=gr(:) ;

if i==1,
  h=gr ; g=gr ;
else
  gamma=(gr-g)'*gr/(g'*g) ;
  g=gr ; h=gamma*h+g ;
end ;

[x J]=make_step(J0,-h,x,critfun) ; 
disp(['Iteration ' num2str(i) ' J=' num2str(J)]) ;
if i>miniter & (J0-J)/(J+1e-50)<prec,  break ; end ;
end ;

% ---------------------------------------------------------------
function [newx,newJ]=make_step(J0,h,x,critfun)
% This function find minimum of J along the direction h
% using simplified Brent's method

maxiniter=500 ;% maximum number of inner loop iterations
sml=1e-4 ;    % small number :-)
gld=0.38197 ; % 1-golden ratio
tol=1e-4 ;    % relative error
abstol=1e-20 ; % absolute error

alpha0=0 ;

magh=sqrt(sum(h.^2))/10 ; % expect approximately unitary steps
if magh>1, h=h/magh ; end ;

% make full step first
J1=feval(critfun,x+h) ;
stps=2 ;
alpha1=1 ;     % save the full step result
if J1<J0,
  J=J0 ;
  % the function seems to decrease, proceed
  while J1<J, 
    J=J1 ; alpha=alpha1 ;
    alpha1=10*alpha1 ;
    J1=feval(critfun,x+h*alpha1) ;
    stps=stps+1 ;
  end ;
else
  % the function increases, go bacward until it decreases
  alpha=0.5 ;
  J=feval(critfun,x+h*alpha) ; 
  while J>J0,
    J1=J ; alpha1=alpha ;
    alpha=0.1*alpha1 ;
    J=feval(critfun,x+h*alpha) ; 
    stps=stps+1 ;
  end ;  
end ;

% now the minimum is bracketed between alpha0,alpha, alpha1
for j=1:maxiniter,
  if j>1 & abs(alpha1-alpha0)<max([tol*alpha abstol]), % success
    newx=x+h*alpha ;
    disp(['Convergence in ' num2str(j) ' + ' num2str(stps) ' steps. ' ...
	  'alpha=' num2str(alpha)]) ;
    newJ=J ;
    return ;
  end ;
  % find new alpha by parabolic approximation
  dif1=alpha-alpha1 ; dif2=J-J0 ; dif3=alpha-alpha0 ; dif4=J-J1 ;
  den=dif1*dif2-dif3*dif4 ;
  if abs(den)<eps, % colinear points
    if J1>J0,
      newalpha=alpha*(1-gld)+gld*alpha0 ;
    else 
      newalpha=alpha1*gld+(1-gld)*alpha ;
    end ;
  else
    newalpha=alpha-0.5*(dif1^2*dif2-dif3^2*dif4)/den ;
    % force into bounds
    newalpha=max([min([newalpha alpha1-sml*(alpha1-alpha0)]) ...
	  sml*(alpha1-alpha0)+alpha0]) ;
  end ;  
  % we have the new alpha now
  newJ=feval(critfun,x+h*newalpha) ; 

  if newJ>J, % something's gone wrong - unsuccessful estimate
    if newalpha<alpha,
      alpha0=newalpha ; J0=newJ ;
    else
      alpha1=newalpha ; J1=newJ ;
    end ;
  else
    if newalpha<alpha ;
      alpha1=alpha ; J1=J ; alpha=newalpha ; J=newJ ;
    else
      alpha0=alpha ; J0=J ; alpha=newalpha ; J=newJ ;
    end ;
  end ;
  
end ; % of for j

disp('Divergence in make_step') ;  
newx=x+h*alpha ;
newJ=J ;

%----------------------------------------------------------------------
