/*		mex_filiirs.c	: Matlab .MEX file
 
 M. Unser / AUG-93 

  The matlab calling syntax is:

			[y] = filiirs(x, zi, c0)

Modified by J.K. April 1998, to compile under Matlab 5 and gcc

*/

#include <math.h>
#include "mex.h"

/*************************************************************
 *	Title : filiirs.c
 *
 * PURPOSE : Symmetric infinite impulse response filter
 * of the form :
 *		H[z] = c0 * Product[ -zi[i] / ( (1-zi[i]*z) (1-zi[i]/z) ), {i,1,nz}]
 *
 * Signals are extended using mirror boundary conditions.
 *
 *  Michael Unser / BEIP	JAN-90 / AUG-93
 *
 * Input signal  : x[k], k=0,....,nx-1
 * Output signal : y[k], k=0,....,nx-1
 *
 * Note : For a unity gain, c0 should be set to
 * c0= Product[(1-zi[i])^2/-zi[i],{i,1,nz}]
 *************************************************************/
#define MIN(x,y) ( ( (x) < (y) ) ? (x) : (y) )

void filiirs(double x[],double y[],int nx,double zi[],int nz,double c0);
void filiirs(double x[],double y[],int nx,double zi[],int nz,double c0)
{
	double x0,xN,log10_epsi=-8.,xk;
	int k,iz,k0;
	   for (k=0;k<nx;k++) y[k]=c0*x[k];
	   for (iz=0;iz<nz;iz++)
	   {
	        k0=(int)(log10_epsi/log10(fabs(zi[iz])));
	        k0=MIN(k0,nx-1);
//  mex_printf("k0=%d" ,(INT)k0);
		x0=y[k0];xN=y[nx-1];

		// Here is a problem if k0>nr
		for (k=k0-1;k>=0;k--) x0=y[k]+zi[iz]*x0;
		y[0]=x0;
		for (k=1;k<nx;k++) y[k]=y[k]+zi[iz]*y[k-1];
		y[nx-1]=-zi[iz]*(2.*y[nx-1] - xN) / (1.-zi[iz]*zi[iz]);
		for (k=nx-2;k>=0;k--) y[k]=zi[iz]*(y[k+1]-y[k]);
	   }

}

/************************************************************************/
/* +=============================+										*/
/* | FUNCTION: iirConvolveMirror |										*/
/* +=============================+										*/
/*----------------------------------------------------------------------*/
/*																		*/
/* Objective:	Discrete recursive convolution							*/
/*																		*/
/* Conventions:	The kernel has a size np								*/
/*				The kernel is given by its (real) poles					*/
/*				The signal has an infinite, periodic (2 n - 1) support	*/
/*				The input is described by n values; the missing values	*/
/*					are provided by mirror symetry						*/
/*				The output has same length n than the input				*/
/*				The output has already been allocated					*/
/*																		*/
/* Comment:		The output may share data storage with the input		*/
/*																		*/
/************************************************************************/
void				iirConvolveMirror
									(double				*inPtr,
									 double				*outPtr,
									 long				n,
									 double				gain,
									 double				*realPoles,
									 long				np) {

  double				*p, *q, *r;
  double				tolerance = log10((double)FLT_EPSILON);
  double				x0, pole;
  long				n2;
  long				i, j, k;
  
  p = inPtr;
  q = outPtr;
  r = outPtr + (ptrdiff_t)n;
  while (q < r)
    *q++ = *p++ * gain;
  if (n == 1L) {
    p = realPoles;
    q = realPoles + (ptrdiff_t)np;
    while (p <= --q)
      *outPtr /= (1.0 - *q) * (1.0 - *q);
    return;
  }
  n2 = 2L * (n - 1L);
  for (i = -np; (i < 0L); i++) {
    pole = *realPoles++;
    j = (long)ceil(tolerance / log10(fabs(pole)));
    k = j - n2 * (j / n2);
    j -= k;
    if (k < n) {
      p = outPtr + (ptrdiff_t)k;
      x0 = *p;
    }
    else {
      k = n2 - k;
      p = outPtr + (ptrdiff_t)k;
      x0 = *p;
      while (++p < r)
	x0 = pole * x0 + *p;
      p--;
    }
    while (--p >= outPtr)
      x0 = pole * x0 + *p;
    while (j > 0L) {
      p++;
      while (++p < r)
	x0 = pole * x0 + *p;
      p--;
      while (--p >= outPtr)
	x0 = pole * x0 + *p;
      j -= n2;
    }
    q = p++;
    *p++ = x0;
    x0 = *(q++ + (ptrdiff_t)n);
    while (p < r)
      *p++ += *q++ * pole;
    *q = (2.0 * *q - x0) / (1.0 - pole * pole);
    p--;
    while (--q >= outPtr)
      *q += *p-- * pole;
  }
} /* End of iirConvolveMirror */



/* Input Arguments */

#define	X_IN	prhs[0]
#define	Z_IN	prhs[1]
#define	C0_IN	prhs[2]

/* Output Arguments */

#define	Y_OUT	plhs[0]


#define	max(A, B)	((A) > (B) ? (A) : (B))
#define	min(A, B)	((A) < (B) ? (A) : (B))


#ifdef __STDC__
void mexFunction(
	int		nlhs,
	mxArray	*plhs[],
	int		nrhs,
	const mxArray	*prhs[]
	)
#else
mexFunction(nlhs, plhs, nrhs, prhs)
int nlhs, nrhs;
mxArray *plhs[], *prhs[];
#endif
{
	double	*y;
	double	*x,*zi,c0;
	unsigned int	m,n;
	int nx,nz,nmin;


	/* Check for proper number of arguments */

	if (nrhs != 3) {
		mexErrMsgTxt("FILIIRS requires three input arguments.");
	} else if (nlhs != 1) {
		mexErrMsgTxt("FILIIRS requires one output argument.");
	}


	/* Check the dimensions of input variables */

	m = mxGetM(X_IN);
	n = mxGetN(X_IN);
	nx=max(m,n);nmin=min(m,n);

	if (!mxIsNumeric(X_IN) || mxIsComplex(X_IN) || 
		mxIsSparse(X_IN)  || !mxIsDouble(X_IN) ||
		(nmin != 1)) {
		mexErrMsgTxt("FILIIRS requires that X be a vector.");
	}

	/* Create a matrix for the return argument */

	Y_OUT = mxCreateDoubleMatrix(m,n,mxREAL) ;

	m = mxGetM(Z_IN);
	n = mxGetN(Z_IN);
	nz=max(m,n);nmin=min(m,n);

	if (!mxIsNumeric(Z_IN) || mxIsComplex(Z_IN) || 
		mxIsSparse(Z_IN)  || !mxIsDouble(Z_IN) ||
		(nmin != 1)) {
		mexErrMsgTxt("FILIIRS requires that ZI be a vector.");
	}

	/* Assign pointers to the various parameters */

	y = mxGetPr(Y_OUT);
	x = mxGetPr(X_IN);
	zi  = mxGetPr(Z_IN);
	c0= (double) *mxGetPr(C0_IN);
	
//	mex_printf("nx=%d, nz=%d, c0=%f",(INT) nx,(INT) nz, (DOUBLE) c0);

	/* Do the actual computations in a subroutine */
#if 0
	filiirs(x,y,nx,zi,nz,c0);
#else
	iirConvolveMirror(x,y,nx,c0,zi,nz) ; /* P.T. */
#endif
	return;
}

