/*
  imwarpcalcreg    The C source file, routine called from meximwarpcalc

% Given a reference image fr, test image ft, its cubic B-spline
% interpolation coefficients ftb and warp coefficient matrix c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g that is monotonous in both x and y
% Now adds penalty function
% Generalized for degrees 1,2,3. Added regularization Tikhonov stabilizer.
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns. The warping matrix is stored columnwise in a vector, 
% the number of columns given by nc
%


  Jan Kybic, May 1998

*/


/* Things to set:      ------------------------------------------ */

#define ALPHA (30.0)   /* exponent multiplicative constant for the penalty */
#define PENDIST (10.0) /* penalty is considered zero for exponent less
			  than -PENDIST */
#define GAMMAK (1.0)  /* normalized penalty per pixel */

/* end of things to set ----------------------------------------- */

#include <math.h>
#ifndef MAKING_MEX
#include <stdio.h>
#include <stdlib.h>
#define mexErrMsgTxt(x) fprintf(stderr,x) 
#define NAME imwarpcalcreg
#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)malloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEFREE(x) mxFree(x)
#endif /* ifndef MAKING_MEX */

#define MAXSUPPORT 4 /* maximal support length */

#define PROFILING 0 /* measure time */

#if PROFILING
#include <sys/param.h>
#include <sys/types.h>
#include <sys/times.h>
#include <limits.h>
#endif

static void intmultbspln(double lambda,double *c,int i,int j,int k,int nd,double *E,double *gr) ;
/* prototype needed, description later */


/* c is the 2D matrix of coefficients decribing the warping function
   by using splines of degree `degree', with nc columns
   and nd rows. fr is the reference image with nr columns and ny rows,
   ft is the test image with nt columns and ny rows, while ftb is a
   matrix of cubic B-spline coefficients for interpolating ft. The
   criterion is returned in E, its gradient in gr and Hessian in
   H. All matrices are stored columnwise, i.e. in the Matlab/Fortran
   fashion. lambda is the regularization factor for a Tikhonov stabilizer. */

void imwarpcalcreg(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny,int degree,
		   double lambda) 
{
  double *e,*fwt,*dereg,*q,tmp,der,r,beta,alpha,gamma,aa ;
  double maxf,parx,pary,betay ;
  int i,j,k,ii,ll,lh,cl,ch,l,lly,lhy,iy,iiy,ind1,ind2 ;
  double tmp1,tmp2,tmp3,tmp4 ;
  double dder ;
  double tabx[MAXSUPPORT],taby[MAXSUPPORT] ;
  double dtabx[MAXSUPPORT+1] ;
  double lastr,qq,lastparx ;
  int lastll, support ;
  double (*bspln)(), (*binterpd)(), shift ;
  void (*imwarp)() ;

  #if PROFILING
  struct tms tms1,tms2 ;
  #endif

  /* set the pointers to the correct functions according to degree */
  switch(degree) {
  case 1: bspln=lbspln ; imwarp=imwarpldd ; binterpd=lbinterpd ;
          break ;
  case 2: bspln=qbspln ; imwarp=imwarpbddh ; binterpd=qbinterpd ;
          break ;
  case 3: bspln=cbspln ; imwarp=imwarpcdd ; binterpd=cbinterpd ;
          break ;
  default: mexErrMsgTxt(NAME " unsupported degree") ;
           return ;
  }
  
  support=degree+1 ;
  shift=(degree-1)/2.0 ;

  if (support>MAXSUPPORT) {
    mexErrMsgTxt(NAME " increase MAXSUPPORT.") ;
    return ;
  }

  /* allocate temporary memory needed */
  SAFEMALLOC(e,ny,nr) ;  SAFEMALLOC(fwt,ny,nr) ;
  SAFEMALLOC(dereg,1,nc*nd) ;
  SAFEMALLOC(q,1,nc-1) ;

  /* warp the test image - always Matlab ordering*/  
  imwarp(fwt,ftb,c,nr,nt,nc,nd,ny) ;
  /* calculate differences and amplitude of the images */
  *E=0.0 ; maxf=0.0 ;
  for(i=0;i<nr;i++) /* columns */
    for(j=0;j<ny;j++) /* rows */
      { tmp1=fr[j+ny*i] ; tmp=tmp1-fwt[j+ny*i] ; 
        e[j+ny*i]=tmp ; (*E)+=tmp*tmp ; 
        tmp1*=tmp1 ;
        if (tmp1>maxf) maxf=tmp1 ;
      }

  /* zero the elements of the gradient and Hessian */
  for(i=0;i<nc*nd;i++) { 
    gr[i]=0.0 ;
    for(j=0;j<nc*nd;j++) H[j*nc+i]=0.0 ;
  }

#define CGRIY(x) gr[iy+nd*(x)]
#define HXY(y,x) H[(y)+(x)*nc*nd]

  /* constants for the penalty function */
  alpha=ALPHA ; gamma=maxf*nr*ny*GAMMAK ; aa=alpha*alpha ;

#if PROFILING
  times(&tms1) ;
#endif

  /* now cycle through all the rows */
  for(j=0;j<ny;j++) {
    pary=1+shift+((double)j)/(ny-1)*(nd-1-2*shift) ;
    for(k=0;k<nr;k++) { /* cycle through columns */
      parx=1+shift+((double)k)/(nr-1)*(nc-1-2*shift) ;
      r=binterpd(c,nc,nd,parx,pary) ;
      /* r now contains g(x) - the x coordinate in the test image */
      /* find the first and last influential parameter, counting from 0 ! */
      /* Influential c are then (ll \to lh)\times(lly \to lhy) */
      ll=(int)floor(parx-shift-1) ; lh=ll+support-1 ;
      if (ll<0) ll=0 ; if (lh>=nc) lh=nc-1 ;
      lly=(int)floor(pary-shift-1) ; lhy=lly+support-1 ;
      if (lly<0) lly=0 ; if (lhy>=nd) lhy=nd-1 ;

      /* determine influential columns counting from 0 */
      cl=(int)floor(r)-2 ; ch=cl+3 ;
      /* The magic constants 2 and 3 are here because the interpolating spline
	 for the image is cubic */
      if (cl<0) cl=0 ; if (ch>=nr) ch=nr-1 ;

      /* calculate first and second derivative of ft at the given coordinates */
      for(der=0.0,l=cl;l<=ch;l++) der-=ftb[j+ny*l]*cbderiv(r-l-1) ;
      for(dder=0.0,l=cl;l<=ch;l++) dder-=ftb[j+ny*l]*cbdderiv(r-l-1) ;

      /* loop over influential parameters */
      for(i=ll;i<=lh;i++) {
	beta=tabx[i-ll]=bspln(parx-i-1) ;	
	for (ii=lly;ii<=lhy;ii++) {
	  betay=taby[ii-lly]=bspln(pary-ii-1) ;
	  gr[ii+i*nd]+=2.0*(dereg[ii+i*nd]=der*beta*betay)*e[j+ny*k] ;
	}
      }

      /* let us deal with the penalty function */
#if 1 
      /* no penalty */
      if (k>0) {
	qq=alpha*(lastr-r) ;
	if (qq>-PENDIST) {
	  (*E)+= (qq=gamma*exp(qq)) ;
	  qq*=alpha ;
	  /* precompute the difference between B-splines */
	  for(i=lastll;i<=lh;i++) 
	    dtabx[i-lastll]=bspln(lastparx-i-1)-bspln(parx-i-1) ;
	  for(i=lastll;i<=lh;i++) {
	    tmp=qq*dtabx[i-lastll] ;
	    for (iy=lly;iy<=lhy;iy++) {
	      betay=tmp*taby[iy-lly] ;
	      CGRIY(i)+=betay ;
	      /* Incorporate penalty into the Hess matrix */

	      tmp3=betay*alpha ;
	      ind1=i*nd+iy ;
	      for(ii=lastll;ii<=lh;ii++) {
		tmp1=tmp3*dtabx[ii-lastll] ;
		for(iiy=lly;iiy<=lhy;iiy++) {
		    ind2=ii*nd+iiy ;
		    tmp2=tmp1*taby[iiy-lly] ;
		    HXY(ind1,ind2)+=tmp2 ;
		  } /* for iiy */
	      } /* for iy */
	    } /* for iy */
	  } /* for i */
	} /* if qq>-PENDIST */
      } /* if k>0 */
#endif
      lastll=ll ;
      lastr=r ;
      lastparx=parx ;

      tmp3=2.0*dder*e[j+ny*k] ;

      /* Update Hess matrix */
      for(i=ll;i<=lh;i++) {
	tmp4=tmp3*tabx[i-ll] ;
	for(iy=lly;iy<=lhy;iy++) {
	  ind1=i*nd+iy ;
	  tmp1=tmp4*taby[iy-lly] ;
	  for(ii=ll;ii<=lh;ii++) {
	    tmp2=tmp1*tabx[ii-ll] ;
	    for(iiy=lly;iiy<=lhy;iiy++) {
	       ind2=ii*nd+iiy ;
		tmp=2.0*dereg[ind1]*dereg[ind2] ; 
		HXY(ind1,ind2)+=tmp ; 
		tmp=tmp2*taby[iiy-lly] ;
		HXY(ind1,ind2)+=tmp ; 
	    } /* for iiy */
	  } /* for ii */ 
	} }/* for iy and for i */
    } /* end of the loop across columns */
  }/* end of the loop across rows */
#if PROFILING
      times(&tms2) ;
      mexPrintf("First loop took %g s.\n",
		((double)tms2.tms_utime-tms1.tms_utime)/CLK_TCK) ;
#endif

  SAFEFREE(e) ; SAFEFREE(fwt) ; SAFEFREE(dereg) ; SAFEFREE(q) ;

  /* here we shall add the Tikhonov stabilizer regularization factor */
  
  /* so far, it is not implemented for other degrees than 3 */
  if (degree!=3) {
    mexErrMsgTxt(NAME 
		 "Tikhonov regularization is implemented for degree=3 only.") ;
    return ; }

   
  /* calculate the value of the criterion and the gradient. 
     Changes to Hessian are neglected for the moment. */
#if 0
  mexPrintf("imwarpcalcreg, lambda=%g E=%g\n",lambda,*E) ; 
#endif

  lambda*=maxf*nr*ny/(nc-2)/(nd-2) ; 
  /* normalize lambda according to the number of coefs. */ 

#if 0
  for(j=1;j<nd-2;j++) { /* for all rows except the first and last */
    for(i=3;i<nc-3;i++) intmultbspln(lambda*8.0/3.0,c,i,i,j,nd,E,gr) ;
    for(i=2;i<nc-3;i++) intmultbspln(-3.0*lambda,c,i,i+1,j,nd,E,gr) ;
    for(i=0;i<nc-3;i++) intmultbspln(lambda/3,c,i,i+3,j,nd,E,gr) ;
    intmultbspln(lambda/3.0,c,0,0,j,nd,E,gr) ;
    intmultbspln((nc>4 ? 4.0/3.0 : 1.0)*lambda,c,1,1,j,nd,E,gr) ;
    intmultbspln((nc>5 ? 7.0/3.0 : (nc==5  ? 2.0:1.0))*lambda,c,2,2,j,nd,E,gr) ;
    if (nc>5) intmultbspln(7.0/3.0*lambda,c,nc-3,nc-3,j,nd,E,gr) ;
    if (nc>4) intmultbspln(4.0*lambda/3.0,c,nc-2,nc-2,j,nd,E,gr) ;
    intmultbspln(lambda/3.0,c,nc-1,nc-1,j,nd,E,gr) ;
    intmultbspln(-lambda,c,0,1,j,nd,E,gr) ;
    intmultbspln(-(nc>4 ? 2.0 : 1.0)*lambda,c,1,2,j,nd,E,gr) ;
    if (nc>4) intmultbspln(-(nc>5 ? 2.0 : 1.0)*lambda,c,nc-3,nc-2,j,nd,E,gr) ;
    intmultbspln(-lambda,c,nc-2,nc-1,j,nd,E,gr) ;
  } /* end of for j */
#endif
#if PROFILING
  times(&tms1) ;
#endif


  for(j=1;j<nd-2;j++) { /* for all rows except the first and last */
    for(i=1;i<nc-2;i++) {
      intmultbspln(lambda/3.0,c,i-1,i-1,j,nd,E,gr) ;
      intmultbspln(-lambda,c,i-1,i,j,nd,E,gr) ;
      intmultbspln(lambda/3.0,c,i-1,i+2,j,nd,E,gr) ;
      intmultbspln(lambda,c,i,i,j,nd,E,gr) ;
      intmultbspln(-lambda,c,i,i+1,j,nd,E,gr) ;
      intmultbspln(lambda,c,i+1,i+1,j,nd,E,gr) ;
      intmultbspln(-lambda,c,i+1,i+2,j,nd,E,gr) ;
      intmultbspln(lambda/3.0,c,i+2,i+2,j,nd,E,gr) ;
    } /* for i */
  } /* for j */

#if PROFILING
      times(&tms2) ;
      mexPrintf("Second loop took %g s.\n",
		((double)tms2.tms_utime-tms1.tms_utime)/CLK_TCK) ;
#endif

#if 0
 mexPrintf("imwarpcalcreg-post, E=%g\n",*E) ; 
#endif
} /* end of imwarpcalcreg */

static void intmultbspln(double lambda,double *c,int i,int j,int k,int nd,
			 double *E,double *gr) 
/* c is stored in Fortran/Matlab ordering. c[row,column]=c[row+column*nd]
   let f1(x)=c[k-1,i]*B(x+1)+c[k,i]*B(x)+c[k+1,i]*B(x-1)+c[k+2,i]*B(x-2) and
       f2(x)=c[k-1,j]*B(x+1)+c[k,j]*B(x)+c[k+1,j]*B(x-1)+c[k+2,j]*B(x-2), 
   where B() is a centered cubic B-spline.
   It calculates  lambda\int_0^1 f1(x) f2(x) dx and ads it to *E, also updates *gr with 
   the respective derivatives of the regularization criterion with respect to c's.
*/

{ static const double Z[4][4]={{20,129,60,1},{129,1188,933,60},
			       {60,933,1188,129},{1,60,129,20}} ;
  double sum,l ;
  int q,w ;

  l=lambda/5040.0 ;

#if 0 
  mexPrintf("intmultbspln called, i=%d,j=%d,k=%d,nd=%d\n",i,j,k,nd) ;
#endif

  k-- ;
  for (sum=0.0,q=0;q<4;q++) 
    for (w=0;w<4;w++) { 
      sum+=Z[q][w]*c[k+w+j*nd]*c[k+q+i*nd] ;
      gr[k+w+j*nd]+=l*Z[q][w]*c[k+q+i*nd] ;
      gr[k+q+j*nd]+=l*Z[q][w]*c[k+w+i*nd] ;
    }

  *E+=l*sum ;

#if 0
 mexPrintf("intmultbspln returned\n") ;
#endif 
}
