/*
  meximwarpcalc    The MEX source file

% Given a reference image fr, test image ft, its cubic B-spline
% interpolation coefficients ftb and warp coefficient matrix c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g that is monotonous in both x and y
% The order of approximation for g is given by (order)
% Uses penalty function based on per-pixel checking of g
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns. The warping matrix is stored columnwise in a vector, 
% the number of columns given by nc
%
% lambda is a coefficient for the regularization Tikhonov stabilizer, 
% so far implemented only  for degree 3.
%
% Optimised and generalised version of meximwarpcalcbddh.c


  Usage: [E,gr,H]=meximwarpcalc(c,fr,ft,ftb,nc,degree,lambda)

  Jan Kybic, June 1998

*/

#define myCreateMatrix(a,b) mxCreateDoubleMatrix(a,b,mxREAL)
#define myAlloc(a) mxMalloc(a)

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt(NAME "IMWARPCALCBDDH: could not malloc") ;

#define SAFEFREE(x) mxFree(x)

#include <string.h>
#include <math.h>
#include "mex.h"

/* Input Arguments */

#define	C_IN	  prhs[0]
#define	FR_IN	  prhs[1]
#define	FT_IN	  prhs[2]
#define	FTB_IN	  prhs[3]
#define NC_IN     prhs[4]
#define DEGREE_IN  prhs[5]
#define LAMBDA_IN prhs[6]

/* Output Arguments */

#define	E_OUT	plhs[0]
#define	GR_OUT	plhs[1]
#define	H_OUT	plhs[2]

#define NAME "MEXIMWARPCALC"

#define max(a,b) ((a>b) ? (a) : (b))

void imwarpcalc(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny,int degree) ;
void imwarpcalcreg(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny,int degree,double lambda) ;

void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
{ double *c,*fr,*ft,*ftb,*E,*gr,*H,lambda ;
  int i,nc,ny,nr,nt,nd,nct,degree ;

  /* Check for proper number of arguments */
  
  if (nrhs != 7) {
    mexErrMsgTxt(NAME " requires seven input arguments.");
  } else if (nlhs != 3) {
    mexErrMsgTxt(NAME " requires three output argument.");
  }
  
  for (i=0;i<nrhs;i++)
    { if (!mxIsNumeric(prhs[i]) || mxIsComplex(prhs[i]) || 
      mxIsSparse(prhs[i])  || !mxIsDouble(prhs[i])) 
         mexErrMsgTxt(
             NAME " needs all input arguments to be real.");
     } 
  
  ny=mxGetM(FR_IN) ; nr=mxGetN(FR_IN) ; nt=mxGetN(FT_IN) ;
  nc=(int)(*(mxGetPr(NC_IN))) ;
  nct=max(mxGetM(C_IN),mxGetN(C_IN)) ;
  nd=nct/nc ;  
  degree=(int)(*(mxGetPr(DEGREE_IN))) ;
  lambda=(double)(*(mxGetPr(LAMBDA_IN))) ;

  if (ny!=mxGetM(FT_IN) || ny!=mxGetM(FTB_IN) || nt!=mxGetN(FTB_IN))
    mexErrMsgTxt(NAME " - wrong size of arguments.") ;

  fr = mxGetPr(FR_IN); ft=mxGetPr(FT_IN) ; c=mxGetPr(C_IN); 
  ftb=mxGetPr(FTB_IN); 

  E_OUT = myCreateMatrix(1,1); E=mxGetPr(E_OUT) ;
  GR_OUT = myCreateMatrix(1,nct); gr=mxGetPr(GR_OUT) ;
  H_OUT = myCreateMatrix(nct,nct); H=mxGetPr(H_OUT) ;

  /*  %%%%% here the calculation begins %%%%%% */
  if (degree!=3 || lambda<1e-30)
    imwarpcalc(c,fr,ft,ftb,E,gr,H,nr,nt,nc,nd,ny,degree) ; 
  else
    imwarpcalcreg(c,fr,ft,ftb,E,gr,H,nr,nt,nc,nd,ny,degree,lambda) ;

}


#define MAKING_MEX

#include "../spline_c/bsplinetools.c"

/* Include both functions here so that we can choose */
#include "imwarpcalc.c"
#include "imwarpcalcreg.c"

/* end of meximwarpcalc */
