function [c,E,nca]=pimwarpopt(frp,ftp,nc,nd,degree) ;
% Given a test image ft and reference image fr, find warp coefficients c
% so that the total square difference E between the reference image and
% warped test image is minimal
% nc is the number of nodes interpolating the warping function horizontally
% and nd vertically 
%
% degree gives degree of the warping function
%
% Uses Levenberg-Marquardt algorithm inspired by Numerical Recipies
% Does not check for constraints but uses penalty function instead
% Now uses quadratic B-spline interpolation for the warping function g
% Multiresolution pyramid approach
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns. In nca it returns the actual final number of columns
% for c, number of rows must be deduced from the size of c
% 
% Usage: [c,E,nca]=pimwarpopt(frp,ftp,nc,nd,degree) ;


fr=frp{1} ; ft=ftp{1} ;
% Calculate the pyramides
nl=min(log(length(fr)),log(length(ft)))/log(2)-4 ;
nl=floor(max(nl,0)) ;
%disp('Calculating pyramides') ;
%figure(1) ;
%frp=pyramid(fr,nl) ; 
%figure(2) ;
%ftp=pyramid(ft,nl) ;

imaging=1 ;
dispwfflg=0 ; % show warping function


if degree==3,
  nca=4 ; nda=4 ;
elseif degree==2,
  nca=4 ; nda=4 ;
elseif degree==1,
  nca=2 ; nda=2 ;
else
  error('Unsupported degree for B-spline warping function') ;
end ;

cmsf=64 ; % number of colours

[ony onr]=size(fr) ; 
[fullny fullnr]=size(fr) ;

for i=nl:-1:1,
  frc=frp{i} ; ftc=ftp{i} ;
  if i~=nl, 
    oldnr=nr ; 
  end ;
  ny=size(frc,1) ; nr=size(frc,2) ; nt=size(ftc,2) ;
  disp([ 'Level ' num2str(i) ' rows=' num2str(ny) ' colref=' num2str(nr) ...
	' coltest=' num2str(nt) ' nca=' num2str(nca) ' nda=' num2str(nda)]) ;

if imaging,  
  subplot(221) ; image(cmsf*frc) ; title('reference') ; colorbar ; 
  subplot(222) ; image(cmsf*ftc) ; title('test') ; colorbar ;
  colormap(jet) ;
  pause(0.01) ;  
end ;  

  if i==nl, % The first time -> initialize c
    if degree==3,
      cr=clincf(1,nr) ;
      c=zeros(4,4) ;
      c(:,1)=clincf(cr(1),cr(1)) ;
      c(:,2)=clincf(cr(2),cr(2)) ;
      c(:,3)=clincf(cr(3),cr(3)) ;
      c(:,4)=clincf(cr(4),cr(4)) ;
      c=reshape(c,1,nca*nda) ;
    elseif degree==2,
      cr=qlincf(1,nr) ;
      c=zeros(4,4) ;
      c(:,1)=qlincf(cr(1),cr(1)) ;
      c(:,2)=qlincf(cr(2),cr(2)) ;
      c(:,3)=qlincf(cr(3),cr(3)) ;
      c(:,4)=qlincf(cr(4),cr(4)) ;
      c=reshape(c,1,nca*nda) ;
    else
      % linear case
      cr(1)=1; cr(2)=nr ;
      c=[cr(1) cr(1) cr(2) cr(2)] ;
    end ;
    if dispwfflg, figure ; dispwf(c/nr,nca,degree) ; axis off ; end ;
  else
    % image dimension was increased, change c appropriately
    gamma=(nr-1)/(oldnr-1) ;
    delta=1-gamma ;
    c=delta+gamma*c ;
  end ;


  
  % Optimize 
  [c,E]=imwarpopti(frc,ftc,nca,nda,c,degree) ;
  
    if dispwfflg, figure ; dispwf(c/nr,nca,degree) ; axis off ; end ;
  
  imeb=imcbanal(ftc) ;
  imh=meximwarp(imeb,c,nr,nca,degree) ;

  
if imaging,  
  subplot(223) ; image(cmsf*imh) ; title('warped') ; colorbar ; grid on ;
  subplot(224) ; image(cmsf*((imh-frc)/2+0.5)) ; title('difference') ; colorbar ; grid on ;
  % disp('Press a key') ; 
  pause(0.01) ;
end ;

  ncaold=nca ; ndaold=nda ;
  % Increase nca, if appropriate
  if degree==3,
    ncn=nca*2-3 ;
  elseif degree==2,
    ncn=nca*2-2 ;
  else
    ncn=nca*2-1 ;
  end;
  
  if ncn<=nc & 4*ncn<=nr,
    nca=ncn ; 
    disp([ 'Expanding parameter space horizontally: nca=' num2str(nca)]) ;
    c=reshape(c,nda,ncaold) ;
    c=imrexp(c,degree) ;
    c=reshape(c,1,nda*nca) ;
  end ;  

  % Increase nda, if appropriate
  if degree==3,
    ndn=nda*2-3 ;
  elseif degree==2,
    ndn=nda*2-2 ;
  else
    ndn=nda*2-1 ;
  end;

  if ndn<=nd & 4*ndn<=ny,
    nda=ndn ;
    disp([ 'Expanding parameter space vertically: nda=' num2str(nda)]) ;
    c=reshape(c,ndaold,nca) ;
    c=imrexp(c',degree)' ;
    c=reshape(c,1,nda*nca) ;
  end ;
  
  if ndaold~=nda | ncaold~=nca,
    % Optimize 
      if dispwfflg, figure ; dispwf(c/nr,nca,degree) ; axis off ; end ;
    [c,E]=imwarpopti(frc,ftc,nca,nda,c,degree) ;
    imh=meximwarp(imeb,c,nr,nca,degree) ;

    if imaging,
    subplot(223) ; image(cmsf*imh) ; title('warped') ; colorbar ; grid on ;
    subplot(224) ; image(cmsf*((imh-frc)/2+0.5)) ; title('difference') ; colorbar ; grid on ;
    % disp('Press a key') ; 
    pause(0.01) ;
    end ;
  end ;  
  
end ;

if dispwfflg, figure ; dispwf(c/nr,nca,degree) ; axis off ; end ;


% possibly move into permissible space - works only for linear and quadratic !

if 0,
  ec=reshape(c,nda,nca) ;
  e=max(max(abs(ec)))*1e-6 ;
  for i=1:nda,
    for j=2:nca,
      if ec(i,j)<ec(i,j-1),  ec(i,j)=ec(i,j-1)+e ; end ;
    end ;
  end ;

c=reshape(ec,1,nca*nda) ;
end ; % end of `if' for move into permissible space 

 