/*
  imwarpcalcdeg    The C source file, routine called from meximwarpcalcdeg

% Given a reference image fr, test image ft, its c/q/l B-spline
% interpolation coefficients ftb and warp coefficient matrix c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g that is monotonous in both x and y
% Now adds penalty function
% Changed to use quadratic interpolation for g
% Boundary points of c are not used directly
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns. The warping matrix is stored columnwise in a vector, 
% the number of columns given by nc
%


  Jan Kybic, May 1998

*/

/* Things to set:      ------------------------------------------ */

#define ALPHA (30.0)   /* exponent multiplicative constant for the penalty */
#define PENDIST (10.0) /* penalty is considered zero for exponent less
			  than -PENDIST */
#define GAMMAK (1.0)  /* normalized penalty per pixel */

/* end of things to set ----------------------------------------- */

#include <math.h>
#ifndef MAKING_MEX
#include <stdio.h>
#include <stdlib.h>
#define mexErrMsgTxt(x) fprintf(stderr,x) 
#define NAME imwarpcalc
#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)malloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEFREE(x) mxFree(x)
#endif /* ifndef MAKING_MEX */

#define MAXSUPPORT 4 /* maximal support length */

void imwarpcdddegone(double *yp,double *fp,double *cp,int n,int nf,int nc,int nd,
		     int ny) {
  imwarpcdddeg(yp,fp,cp,n,nf,nc,nd,ny,1) ; 
}

void imwarpcdddegtwo(double *yp,double *fp,double *cp,int n,int nf,int nc,int nd,
		     int ny) {
  imwarpcdddeg(yp,fp,cp,n,nf,nc,nd,ny,2) ; 
}


/* c is the 2D matrix of coefficients decribing the warping function
   by using splines of degree `degree', with nc columns
   and nd rows. fr is the reference image with nr columns and ny rows,
   ft is the test image with nt columns and ny rows, while ftb is a
   matrix of cubic B-spline coefficients for interpolating ft. The
   criterion is returned in E, its gradient in gr and Hessian in
   H. All matrices are stored columnwise, i.e. in the Matlab/Fortran
   fashion. */

void imwarpcalcdeg(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny,int degree,
		   int degreeim) 
{
  double *e,*fwt,*dereg,*q,tmp,der,r,beta,alpha,gamma,aa ;
  double maxf,parx,pary,betay ;
  int i,j,k,ii,ll,lh,cl,ch,l,lly,lhy,iy,iiy,ind1,ind2 ;
  double tmp1,tmp2,tmp3,tmp4 ;
  double dder ;
  double tabx[MAXSUPPORT],taby[MAXSUPPORT] ;
  double dtabx[MAXSUPPORT+1] ;
  double lastr,qq,lastparx ;
  int lastll, support ;
  double (*bspln)(), (*binterpd)(), shift ;
  void (*imwarp)() ;
  

  /* set the pointers to the correct functions according to degree */
  switch(degree) {
  case 1: bspln=lbspln ; imwarp=imwarpldd ; binterpd=lbinterpd ;
          break ;
  case 2: bspln=qbspln ; imwarp=imwarpbddh ; binterpd=qbinterpd ;
          break ;
  case 3: bspln=cbspln ; imwarp=imwarpcdd ; binterpd=cbinterpd ;
          break ;
  default: mexErrMsgTxt(NAME " unsupported degree") ;
           return ;
  }

  if (degreeim!=3 && degree!=3) 
    mexErrMsgTxt(NAME " unsupported degree combination ") ;
           
  switch(degreeim) { /* image spline interpolation degree */
  case 1: imwarp=imwarpcdddegone ; break ;
  case 2: imwarp=imwarpcdddegtwo ; break ;
  case 3: imwarp=imwarpcdd ; break ;
  default: mexErrMsgTxt(NAME " unsupported image degree") ;
           return ;
  }

  support=degree+1 ;
  shift=(degree-1)/2.0 ;

  if (support>MAXSUPPORT) {
    mexErrMsgTxt(NAME " increase MAXSUPPORT.") ;
    return ;
  }

  /* allocate temporary memory needed */
  SAFEMALLOC(e,ny,nr) ;  SAFEMALLOC(fwt,ny,nr) ;
  SAFEMALLOC(dereg,1,nc*nd) ;
  SAFEMALLOC(q,1,nc-1) ;

  /* warp the test image - always Matlab ordering*/  
  imwarp(fwt,ftb,c,nr,nt,nc,nd,ny) ;
  /* calculate differences and amplitude of the images */
  *E=0.0 ; maxf=0.0 ;
  for(i=0;i<nr;i++) /* columns */
    for(j=0;j<ny;j++) /* rows */
      { tmp1=fr[j+ny*i] ; tmp=tmp1-fwt[j+ny*i] ; 
        e[j+ny*i]=tmp ; (*E)+=tmp*tmp ; 
        tmp1*=tmp1 ;
        if (tmp1>maxf) maxf=tmp1 ;
      }

  /* zero the elements of the gradient and Hessian */
  for(i=0;i<nc*nd;i++) { 
    gr[i]=0.0 ;
    for(j=0;j<nc*nd;j++) H[j*nc+i]=0.0 ;
  }

#define CGRIY(x) gr[iy+nd*(x)]
#define HXY(y,x) H[(y)+(x)*nc*nd]

  /* constants for the penalty function */
  alpha=ALPHA ; gamma=maxf*nr*ny*GAMMAK ; aa=alpha*alpha ;

  /* now cycle through all the rows */
  for(j=0;j<ny;j++) {
    pary=1+shift+((double)j)/(ny-1)*(nd-1-2*shift) ;
    for(k=0;k<nr;k++) { /* cycle through columns */
      parx=1+shift+((double)k)/(nr-1)*(nc-1-2*shift) ;
      r=binterpd(c,nc,nd,parx,pary) ;
      /* r now contains g(x) - the x coordinate in the test image */
      /* find the first and last influential parameter, counting from 0 ! */
      /* Influential c are then (ll \to lh)\times(lly \to lhy) */
      ll=(int)floor(parx-shift-1) ; lh=ll+support-1 ;
      if (ll<0) ll=0 ; if (lh>=nc) lh=nc-1 ;
      lly=(int)floor(pary-shift-1) ; lhy=lly+support-1 ;
      if (lly<0) lly=0 ; if (lhy>=nd) lhy=nd-1 ;

      /* determine influential image columns counting from 0!  */
      switch (degreeim) {
      case 3: cl=(int)floor(r)-2 ; ch=cl+3 ; 
	if (cl<0) cl=0 ; if (ch>=nr) ch=nr-1 ;
	/* The magic constants 2 and 3 are here because the interpolating spline
	   for the image is cubic */
	/* calculate first and second derivative of ft at the given coordinates */
	for(der=0.0,l=cl;l<=ch;l++) der-=ftb[j+ny*l]*cbderiv(r-l-1) ;
	for(dder=0.0,l=cl;l<=ch;l++) dder-=ftb[j+ny*l]*cbdderiv(r-l-1) ;
	break ;

      case 2: cl=(int)floor(r-0.5)-1 ; ch=cl+2 ;
	if (cl<0) cl=0 ; if (ch>=nr) ch=nr-1 ; 
	for(der=0.0,l=cl;l<=ch;l++) der-=ftb[j+ny*l]*qbderiv(r-l-1) ;
	for(dder=0.0,l=cl;l<=ch;l++) dder-=ftb[j+ny*l]*qbdderiv(r-l-1) ;
	break ;
      case 1: cl=(int)floor(r)-1 ; ch=cl+1 ; 
	if (cl<0) cl=0 ; if (ch>=nr) ch=nr-1 ;
	for(der=0.0,l=cl;l<=ch;l++) der-=ftb[j+ny*l]*lbderiv(r-l-1) ;
	dder=0.0 ;
	break ; 	
      }

      /* loop over influential parameters */
      for(i=ll;i<=lh;i++) {
	beta=tabx[i-ll]=bspln(parx-i-1) ;	
	for (ii=lly;ii<=lhy;ii++) {
	  betay=taby[ii-lly]=bspln(pary-ii-1) ;
	  gr[ii+i*nd]+=2.0*(dereg[ii+i*nd]=der*beta*betay)*e[j+ny*k] ;
	}
      }

      /* let us deal with the penalty function */
      if (k>0) {
	qq=alpha*(lastr-r) ;
	if (qq>-PENDIST) {
	  (*E)+= (qq=gamma*exp(qq)) ;
	  qq*=alpha ;
	  /* precompute the difference between B-splines */
	  for(i=lastll;i<=lh;i++) 
	    dtabx[i-lastll]=bspln(lastparx-i-1)-bspln(parx-i-1) ;
	  for(i=lastll;i<=lh;i++) {
	    tmp=qq*dtabx[i-lastll] ;
	    for (iy=lly;iy<=lhy;iy++) {
	      betay=tmp*taby[iy-lly] ;
	      CGRIY(i)+=betay ;
	      /* Incorporate penalty into the Hess matrix */
	      tmp3=betay*alpha ;
	      ind1=i*nd+iy ;
	      for(ii=lastll;ii<=lh;ii++) {
		tmp1=tmp3*dtabx[ii-lastll] ;
		for(iiy=lly;iiy<=lhy;iiy++) {
		    ind2=ii*nd+iiy ;
		    tmp2=tmp1*taby[iiy-lly] ;
		    HXY(ind1,ind2)+=tmp2 ;
		  } /* for iiy */
	      } /* for iy */
	    } /* for iy */
	  } /* for i */
	} /* if qq>-PENDIST */
      } /* if k>0 */
      lastll=ll ;
      lastr=r ;
      lastparx=parx ;

      tmp3=2.0*dder*e[j+ny*k] ;

      /* Update Hess matrix */
      for(i=ll;i<=lh;i++) {
	tmp4=tmp3*tabx[i-ll] ;
	for(iy=lly;iy<=lhy;iy++) {
	  ind1=i*nd+iy ;
	  tmp1=tmp4*taby[iy-lly] ;
	  for(ii=ll;ii<=lh;ii++) {
	    tmp2=tmp1*tabx[ii-ll] ;
	    for(iiy=lly;iiy<=lhy;iiy++) {
	       ind2=ii*nd+iiy ;
		tmp=2.0*dereg[ind1]*dereg[ind2] ; 
		HXY(ind1,ind2)+=tmp ; 
		tmp=tmp2*taby[iiy-lly] ;
		HXY(ind1,ind2)+=tmp ; 
	    } /* for iiy */
	  } /* for ii */ 
	} }/* for iy and for i */
    } /* end of the loop across columns */
  }/* end of the loop across rows */

  SAFEFREE(e) ; SAFEFREE(fwt) ; SAFEFREE(dereg) ; SAFEFREE(q) ;
}

