function [c,E]=spwopt(fr,ft,c,s) ;
% Given a test image ft and reference image fr, find warp coefficients c
% so that the total square difference E between the reference image and
% warped test image is minimal
% For description of s, see spwstruct.m
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns
%
% c is the initial estimate of the warping coefficients
%
% Jan Kybic, November 1998
%
% Usage: [c,E]=spwopt(fr,ft,c,s) ;

displaying=s(11)>1 ; % should progress messages be printed

% Tlambda=0 ; % coefficient for the regularization Tikhonov stabilizer
Tlambda=s(8) ;
[nd,nc]=size(c) ;
c=reshape(c,1,nd*nc) ;

[ny nr]=size(fr) ;
[nyt nt]=size(ft) ;

if ny~=nyt, error('fr and ft must have the same number of rows') ; end ;

degree=s(1) ;
if degree==2,
  if nc<4 & nd<4, error('nc,nd must be at least 4.') ; end ;
end ;


ftb=imsanal(ft,s(19)) ;

% [Eold,gr,H]=meximwarpcalcbddh(c,fr,ft,ftb,nc) ;
[Eold,gr,H]=meximwarpcalcdeg(c,fr,ft,ftb,nc,degree,Tlambda,s(19)) ;

E=Eold ;
Eold=2*Eold+1 ;
prec=s(14)   ;
delta=s(15) ;
lambda=s(17) ; % initial lambda
i=1 ; j=0 ;

imh=meximwarpd(ftb,c,nr,nc,degree,s(19)) ;
Ein=sum(sum((imh-fr).^2)) ;

if s(18)==2, % conjugated gradient method
  h=gr ; g=gr ;
end ;  

while (Eold-E>Eold*prec+delta | j<1), 
  % previously while ((Eold-E)/Eold>prec | j<3),
  if displaying,   
    disp([ 'Iteration ' num2str(i) '  criterion=' num2str(E)]) ; 
  end ;

%  if length(gr)>200,
%    keyboard
% end ;
  
  
  if s(18)==1, % Marquardt-Levenberg
    E1=E ;
    M=H+eye(nc*nd)*lambda ; rc=rcond(M) ;
    if rc>s(16),
      ccc=cputime ;
      if nc<20,
	dc=-gr/M' ;
      else  
	dc=-qmr(M,gr',1e-3,100)' ;
	clear M ;
      end ;
      
      if s(11)>2, disp([ 'Linear system solution took ' ...
	    num2str(cputime-ccc) ' s' ]) ; end ;
      cpdc=c+dc ;
      [E,gr1,H1]=meximwarpcalcdeg(cpdc,fr,ft,ftb,nc,degree,Tlambda,s(19)) ;
      
      if displaying,
	disp(['Returned E=' num2str(E)]) ;
      end ;	
      i=i+1 ;
    else % rc>...
	  if displaying,
	disp('Linear equation set close to singular.') ;
      end ;
    end ; % rc>...
	
    if E>E1 | rc<=s(16),
      lambda=s(9)*lambda ; E=E1 ;
      if displaying,
	disp(['Increasing lambda and trying again lambda=' num2str(lambda)]) ;
      end ;  
      j=0 ;
    else
      lambda=lambda/s(9) ;
      c=cpdc ;
      Eold=E1 ; H=H1 ; gr=gr1 ;
      j=j+1 ;
    end ;
  else % Not Marquardt Levenberg
       % conjugated gradient method
    J0=E ; Eold=E ;   
    gamma=(gr-g)*gr'/(g*g') ;
    g=gr ; h=gamma*h+g ;
    % h=gr ; % Just a stupid gradient descent
    i=i+1 ; j=j+1 ;
    convflg=0 ;
    % we go in the direction of -h, until we find a minimum
    % maxiniter=500 ;% maximum number of inner loop iterations
    maxiniter=50 ;% maximum number of inner loop iterations
    sml=1e-4 ;    % small number :-)
    gld=0.38197 ; % 1-golden ratio
    tol=s(14) ;    % relative error
    abstol=s(15) ; % absolute error
    alpha0=0 ;
    h1=-h ;
    magh=sqrt(sum(h1.^2))/10 ; % expect approximately unitary steps
    if magh>1, h1=h1/magh ; end ;
    [J1,gr1,H1]=meximwarpcalcdeg(c+h1,fr,ft,ftb,nc,degree,Tlambda,s(19)) ;
    stps=2 ;
    alpha1=1 ; % save the full step result
    if J1<J0,
      J=J0 ;
      % the function seems to decrease, proceed
      while J1<J, 
	J=J1 ; alpha=alpha1 ;
	alpha1=10*alpha1 ;
	[J1,gr1,H1]=meximwarpcalcdeg(c+h1*alpha1,fr,ft,ftb,nc,degree,Tlambda,s(19)) ; 
	stps=stps+1 ;
      end ;
    else
      % the function increases, go backward until it decreases
      alpha=0.5 ;
      [J,gr1,H1]=meximwarpcalcdeg(c+h1*alpha,fr,ft,ftb,nc,degree,Tlambda,s(19)) ; 
      while J>J0,
	J1=J ; alpha1=alpha ;
	alpha=0.1*alpha1 ;
	[J,gr1,H1]=meximwarpcalcdeg(c+h1*alpha,fr,ft,ftb,nc,degree,Tlambda,s(19)) ; 
	stps=stps+1 ;
      end ;  
    end ;
    % now the minimum is bracketed between alpha0,alpha, alpha1
    for jj=1:maxiniter,
      if jj>1 & abs(alpha1-alpha0)<max([0.1*tol*alpha abstol]), % success
	newc=c+h1*alpha ;
	if s(11)>1,
	  disp(['Convergence in ' num2str(jj) ' + ' num2str(stps) ' steps. ' ...
		  'alpha=' num2str(alpha)]) ;
	end ;
	newJ=J ;
	convflg=1 ;
	break ;
      else % not converging so far
	% find new alpha by parabolic approximation
	dif1=alpha-alpha1 ; dif2=J-J0 ; dif3=alpha-alpha0 ; dif4=J-J1 ;
	den=dif1*dif2-dif3*dif4 ;
	if abs(den)<eps, % colinear points
	  if J1>J0,
	    newalpha=alpha*(1-gld)+gld*alpha0 ;
	  else 
	    newalpha=alpha1*gld+(1-gld)*alpha ;
	  end ;
	else
	  newalpha=alpha-0.5*(dif1^2*dif2-dif3^2*dif4)/den ;
	  % force into bounds
	  newalpha=max([min([newalpha alpha1-sml*(alpha1-alpha0)]) ...
		sml*(alpha1-alpha0)+alpha0]) ;
	end ; % colinear points  
	
	% we have the new alpha now
	[newJ,gr,H1]=meximwarpcalcdeg(c+h1*newalpha,fr,ft,ftb,nc,degree,Tlambda,s(19)) ; 
	if s(11)>2,
	    disp(['At alpha=' num2str(newalpha,20) ' E=' num2str(newJ,20)]) ;
	    disp(['   alpha0=' num2str(alpha0,20) ' alpha1=' num2str(alpha1,20)]) ;
	end ;
	
	if newJ>J, % something's gone wrong - unsuccessful estimate
	  if newalpha<alpha,
	    alpha0=newalpha ; J0=newJ ;
	  else
	    alpha1=newalpha ; J1=newJ ;
	  end ;
	else
	  if newalpha<alpha ;
	    alpha1=alpha ; J1=J ; alpha=newalpha ; J=newJ ;
	  else
	    alpha0=alpha ; J0=J ; alpha=newalpha ; J=newJ ;
	  end ;
	end ; % if newJ>J
      end ; % jj>1 ..
    end ; % of for jj
    if convflg==0,
      if displaying,
	disp('Divergence in make_step') ;  
      end ;
      newc=c+h1*alpha ;
      newJ=J ;
    end ;
    E=newJ; c=newc ;
  end ; % end of conjugated gradient method
end ; % while
      

if displaying,
  disp([ 'Iterations ' num2str(i) '  criterion=' num2str(E)]) ;
end ;

disp([ 'Iterations ' num2str(i) '  rel. criterion=' ...
      num2str(E/size(fr,1)/size(fr,2))]) ;
imh=meximwarpd(ftb,c,nr,nc,degree,s(19)) ;

Eout=sum(sum((imh-fr).^2)) ;
disp([ 'Criterion ' num2str(Ein) ' -> ' num2str(Eout) ]) ;

c=reshape(c,nd,nc) ;
