function [c,E]=spwpyr(frp,ftp,s) ;
% Given a test image pyramid ftp and reference image pyramid frp, 
% find warp coefficients c
% so that the total square difference E between the reference image and
% warped test image is minimal.
% Number of mesh coefficients can be now arbitrary
% 
% For description of s, see spwstruct.m
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns. 
% 
% Jan Kybic, November 1998
%
% Usage: [c,E]=spwpyr(frp,ftp,s) ;


fr=frp{1} ; ft=ftp{1} ;
% the pyramide level count
nl=size(frp,1) ;

imaging=s(12)>0 ;
dispwfflg=s(13)>0 ; % show warping function
wfcnt=1 ;
nc=s(2) ; nd=s(3) ;
degree=s(1) ;

if degree==3,
  nca=4 ; nda=4 ;
elseif degree==2,
  nca=4 ; nda=4 ;
elseif degree==1,
  nca=2 ; nda=2 ;
else
  error('Unsupported degree for B-spline warping function') ;
end ;

cmsf=length(colormap) ;

[ony onr]=size(fr) ; 
[fullny fullnr]=size(fr) ;

for i=nl:-1:1,
  frc=frp{i} ; ftc=ftp{i} ;
  if i~=nl, 
    oldnr=nr ; 
  end ;
  ny=size(frc,1) ; nr=size(frc,2) ; nt=size(ftc,2) ;
  disp([ 'Level ' num2str(i) ' rows=' num2str(ny) ' colref=' num2str(nr) ...
	' coltest=' num2str(nt) ' nca=' num2str(nca) ' nda=' num2str(nda)]) ;

  if imaging,  
    subplot(221) ; image(cmsf*frc) ; title('reference') ; colorbar ; 
    subplot(222) ; image(cmsf*ftc) ; title('test') ; colorbar ;
    colormap(jet) ;
    pause(0.01) ;  
  end ;  

  if i==nl, % The first time -> initialize c
    if degree==3,
      cr=clincf(1,nr) ;
      c=zeros(4,4) ;
      c(:,1)=clincf(cr(1),cr(1)) ;
      c(:,2)=clincf(cr(2),cr(2)) ;
      c(:,3)=clincf(cr(3),cr(3)) ;
      c(:,4)=clincf(cr(4),cr(4)) ;
    elseif degree==2,
      cr=qlincf(1,nr) ;
      c=zeros(4,4) ;
      c(:,1)=qlincf(cr(1),cr(1)) ;
      c(:,2)=qlincf(cr(2),cr(2)) ;
      c(:,3)=qlincf(cr(3),cr(3)) ;
      c(:,4)=qlincf(cr(4),cr(4)) ;
    else
      % linear case
      cr(1)=1; cr(2)=nr ;
      c=[cr(1) cr(2) ; cr(1) cr(2)] ;
    end ;
    if dispwfflg, figure ; dispwcntr(c) ; axis off ; 
      if s(13)>1, eval(['print -depsc demowf' num2str(wfcnt) '.eps']) ; 
      wfcnt=wfcnt+1 ; end ;
      end ;
  else
    % image dimension was increased, change c appropriately
    gamma=(nr-1)/(oldnr-1) ;
    delta=1-gamma ;
    c=delta+gamma*c ;
  end ;


  
  % Optimize 
  [c,E]=spwopt(frc,ftc,c,s) ;
  
  if dispwfflg, figure ; dispwcntr(c) ; axis off ; 
        if s(13)>1, eval(['print -depsc demowf' num2str(wfcnt) '.eps']) ; 
      wfcnt=wfcnt+1 ; end ;
  end ;
  
  imeb=imsanal(ftc,s(19)) ;
  imh=meximwarpd(imeb,reshape(c,1,size(c,1)*size(c,2)),nr,nca,s(1),s(19)) ;

  if imaging,  
    subplot(223) ; image(cmsf*imh) ; title('warped') ; colorbar ; grid on ;
    subplot(224) ; image(cmsf*((imh-frc)/2+0.5)) ; title('difference') ; 
    colorbar ; grid on ;
    % disp('Press a key') ; 
    pause(0.01) ;
  end ;

  ncaold=nca ; ndaold=nda ;
  % Increase nca, if appropriate
  if degree==3,
    ncn=nca*2-3 ;
  elseif degree==2,
    ncn=nca*2-2 ;
  else
    ncn=nca*2-1 ;
  end;
  
  if ncn<=nc & ncn<=nr, % 4*ncn<=nr,
    nca=ncn ; 
    disp([ 'Expanding parameter space horizontally: nca=' num2str(nca)]) ;
    c=imrexp(c,degree) ;
  end ;  

  % Increase nda, if appropriate
  if degree==3,
    ndn=nda*2-3 ;
  elseif degree==2,
    ndn=nda*2-2 ;
  else
    ndn=nda*2-1 ;
  end;

  if ndn<=nd & ndn<=ny, % 4*ndn<=ny,
    nda=ndn ;
    disp([ 'Expanding parameter space vertically: nda=' num2str(nda)]) ;
    c=imrexp(c',degree)' ;
  end ;
  
  if ndaold~=nda | ncaold~=nca,
    % Optimize 
    if dispwfflg, figure ; dispwcntr(c) ; axis off ; 
          if s(13)>1, eval(['print -depsc demowf' num2str(wfcnt) '.eps']) ; 
      wfcnt=wfcnt+1 ; end ;
    end ;
    [c,E]=spwopt(frc,ftc,c,s) ;
    imh=meximwarpd(imeb,reshape(c,1,size(c,1)*size(c,2)),nr,nca,s(1),s(19)) ;

    if imaging,
      subplot(223) ; image(cmsf*imh) ; title('warped') ; colorbar ; grid on ;
      subplot(224) ; image(cmsf*((imh-frc)/2+0.5)) ; title('difference') ; 
      colorbar ; grid on ;
      % disp('Press a key') ; 
      pause(0.01) ;
    end ;
  end ;  % ndaold~=nda ... image size changed
  
end ; % main loop - all pyramid levels 

% Now, if we have not yet reached the desired resolution, go for it at once

% disp([ 'Main loop finished at nca=' num2str(nca) ' nda=' num2str(nda) ]) ; 
% disp([ 'Expanding to nc=' num2str(nc) ' nd=' num2str(nd) ]) ; 

if(nca<nc | nda<nd),
  
  if nca<nc, 
    nca=nc ; 
    disp([ 'Expanding parameter space horizontally: nca=' num2str(nca)]) ;
    c=imrexpg(c,nc,degree) ;
  end ;  

  if nda<nd, % 4*ndn<=ny,
    nda=nd ;
    disp([ 'Expanding parameter space vertically: nda=' num2str(nda)]) ;
    c=imrexpg(c',nd,degree)' ;
  end ;
  
  % Optimize 
  if dispwfflg, figure ; dispwcntr(c) ; axis off ; 
        if s(13)>1, eval(['print -depsc demowf' num2str(wfcnt) '.eps']) ; 
      wfcnt=wfcnt+1 ; end ;
    end ;
  [c,E]=spwopt(frc,ftc,c,s) ;
  imh=meximwarp(imeb,reshape(c,1,size(c,1)*size(c,2)),nr,nca,s(1)) ;

  if imaging,
    subplot(223) ; image(cmsf*imh) ; title('warped') ; colorbar ; grid on ;
    subplot(224) ; image(cmsf*((imh-frc)/2+0.5)) ; title('difference') ; 
    colorbar ; grid on ;
    % disp('Press a key') ; 
    pause(0.01) ;
  end ;
end ;

if dispwfflg, figure ; dispwcntr(c) ; axis off ; 
      if s(13)>1, eval(['print -depsc demowf' num2str(wfcnt) '.eps']) ; 
      wfcnt=wfcnt+1 ; end ;
end ;



 