function [c,E,w]=imwarpoptbddhil(fr,ft,nc,nd,c,order) ;
% Given a test image ft and reference image fr, find warp coefficients c
% so that the total square difference E between the reference image and
% warped test image is minimal
% nc is the number of nodes interpolating the warping function horizontally
% and nd vertically
%
% Uses Levenberg-Marquardt algorithm inspired by Numerical Recipies
% Does not check for constraints but uses penalty function instead
% Now uses quadratic B-spline interpolation for the warping function g
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns
%
% c is the initial estimate of the warping coefficients
% shift 0.5
%
% Usage: [c,E]=imwarpoptbddhi(fr,ft,nc,nd,c,order) ;


[ny nr]=size(fr) ;
[nyt nt]=size(ft) ;

if ny~=nyt, error('fr and ft must have the same number of rows') ; end ;

if order==2,
  if nc<4 & nd<4, error('nc,nd must be at least 4.') ; end ;
end ;


ftb=imcbanal(ft) ;

[Eold,gr,H]=meximwarpcalcbddh(c,fr,ft,ftb,nc) ;
E=Eold ;
w=E ;
Eold=2*Eold+1 ;
prec=1e-4   ;
delta=1e-10 ;
lambda=0.1 ;
i=1 ; j=0 ;

if order==2,
  imh=meximwarpbddh(ftb,c,nr,nc) ;
else
  imh=meximwarpldd(ftb,c,nr,nc) ;
end ;

Ein=sum(sum((imh-fr).^2)) ;

while (Eold-E>Eold*prec+delta), 
  % previously while ((Eold-E)/Eold>prec | j<3),
  % disp([ 'Iteration ' num2str(i) '  criterion=' num2str(E)]) ;

  E1=E ;
  M=H+eye(nc*nd)*lambda ; rc=rcond(M) ;
  if 1, % classical versus SVD
    if rc>1e-10,
      dc=-gr/M' ;
      cpdc=c+dc ;
      [E,gr1,H1]=meximwarpcalcbddh(cpdc,fr,ft,ftb,nc) ;
      % disp(['Returned E' num2str(E)]) ;
      % E
      i=i+1 ;
    else
      % disp('Linear equation set close to singular.') ;
    end ;
  else
    % svd implementation - more robust but apparently unnecessary
    [u,s,v]=svd(H+eye(nc*nd)*lambda) ;
    for j=1:nc*nd, 
      if abs(s(j,j))>1e-6, 
	s(j,j)=1/s(j,j) ;
      else
	s(j,j)=0 ;
      end ;
    end ;
    dc=-gr*v*s*u' ;
    cpdc=c+dc ;
    [E,gr1,H1]=meximwarpcalcbddh(cpdc,fr,ft,ftb,nc) ;
    % disp(['Returned E' num2str(E)]) ;
    % E
    i=i+1 ;
  end ;
    
  if E>E1 | rc<=1e-10,
    lambda=10*lambda ; E=E1 ;
    % disp(['Increasing lambda and trying again lambda=' num2str(lambda)]) ;
    % j=0 ;
  else
    w=[w E] ;
    lambda=lambda/10 ;
    c=cpdc ;
    Eold=E1 ; H=H1 ; gr=gr1 ;
    % j=j+1 ;
  end ;
end ;

% disp([ 'Iterations ' num2str(i) '  criterion=' num2str(E)]) ;

disp([ 'Iterations ' num2str(i) '  rel. criterion=' ...
      num2str(E/size(fr,1)/size(fr,2))]) ;
if order==2,
  imh=meximwarpbddh(ftb,c,nr,nc) ;
else
  imh=meximwarpldd(ftb,c,nr,nc) ;
end ;

Eout=sum(sum((imh-fr).^2)) ;
disp([ 'Criterion ' num2str(Ein) ' -> ' num2str(Eout) ]) ;
