/*
  meximwarpcalcbbdd    The MEX source file

% Given a reference image fr, test image ft, its cubic B-spline
% interpolation coefficients ftb and warp coefficient matrix c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g that is monotonous in both x and y
% No penalty function, going out of the permissible space is regarded as
% failure 
% Changed to use quadratic interpolation for g
% Boundary points of c are not used directly
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns. The warping matrix is stored columnwise in a vector, 
% the number of columns given by nc

  Usage: [E,gr,H]=imwarpcalcbbdd(c,fr,ft,ftb,nc)

  Jan Kybic, April 1997

*/

/* #ifdef __GNUC__ 
 #ifndef linux 
 #define linux 
 #endif 
 #endif */

#ifdef linux
#define myCreateMatrix(a,b) mxCreateFull(a,b,REAL)
#define myAlloc(a) mxCalloc(a,1)
#else /* hp version */
#define myCreateMatrix(a,b) mxCreateDoubleMatrix(a,b,mxREAL)
#define myAlloc(a) mxMalloc(a)
#endif

#define DEBUG 0

#include <string.h>
#include <math.h>
#include "mex.h"

/* Input Arguments */

#define	C_IN	prhs[0]
#define	FR_IN	prhs[1]
#define	FT_IN	prhs[2]
#define	FTB_IN	prhs[3]
#define NC_IN   prhs[4]

/* Output Arguments */

#define	E_OUT	plhs[0]
#define	GR_OUT	plhs[1]
#define	H_OUT	plhs[2]

#define max(a,b) ((a>b) ? (a) : (b))

void imwarpcalcbbdd(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny) ;

#ifdef linux
void mexFunction(
                 int nlhs, Matrix *plhs[],
                 int nrhs, Matrix *prhs[]
		 )
#else
void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
#endif
{ double *c,*fr,*ft,*ftb,*E,*gr,*H ;
  int i,nc,ny,nr,nt,nd,nct ;

  /* Check for proper number of arguments */
  
  if (nrhs != 5) {
    mexErrMsgTxt("IMWARPCALCBBDD requires five input arguments.");
  } else if (nlhs != 3) {
    mexErrMsgTxt("IMWARPCALCBBDD requires three output argument.");
  }
  
  for (i=0;i<nrhs;i++)
    { if (!mxIsNumeric(prhs[i]) || mxIsComplex(prhs[i]) || 
      mxIsSparse(prhs[i])  || !mxIsDouble(prhs[i])) 
         mexErrMsgTxt(
             "IWARPCALCBBDD needs all input arguments to be real.");
     } 
  
  ny=mxGetM(FR_IN) ; nr=mxGetN(FR_IN) ; nt=mxGetN(FT_IN) ;
  nc=(int)(*(mxGetPr(NC_IN))) ;
  nct=max(mxGetM(C_IN),mxGetN(C_IN)) ;
  nd=nct/nc ;  

  if (ny!=mxGetM(FT_IN) || ny!=mxGetM(FTB_IN) || nt!=mxGetN(FTB_IN))
    mexErrMsgTxt("IMWARPCALCBBDD - wrong size of arguments.") ;

  fr = mxGetPr(FR_IN); ft=mxGetPr(FT_IN) ; c=mxGetPr(C_IN); 
  ftb=mxGetPr(FTB_IN); 

  E_OUT = myCreateMatrix(1,1); E=mxGetPr(E_OUT) ;
  GR_OUT = myCreateMatrix(1,nct); gr=mxGetPr(GR_OUT) ;
  H_OUT = myCreateMatrix(nct,nct); H=mxGetPr(H_OUT) ;

  /*  %%%%% here the calculation begins %%%%%% */
#if DEBUG
  mexPrintf("imwarpcalcbbdd called nc=%d nd=%d nr=%d nt=%d ny=%d\n",nc,nd,
	    nr,nt,ny) ;
#endif 
  imwarpcalcbbdd(c,fr,ft,ftb,E,gr,H,nr,nt,nc,nd,ny) ;
}

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt("IMWARPCALCBBDD: could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt("IMWARPCALCBBDD: could not malloc") ;

#define MAKING_MEX

#include "bsplinetools.c"

/* This is the executive function */

#define CJ(x) c[j+nd*(x)]
#define CGRJ(x) gr[j+nd*(x)]
#define HXY(y,x) H[(y)+(x)*nc*nd]
#define HCJ(y,x) HXY(j+nd*(y),j+nd*(x))

void imwarpcalcbbdd(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny) 
{
  double *e,*fwt,*dereg,*q,tmp,der,r,beta,alpha,gamma,aa ;
  double maxf,parx,pary,betay ;
  int i,j,k,ii,ll,lh,cl,ch,l,lly,lhy,iy,iiy,ind1,ind2 ;

  /* check the contraints */
#if DEBUG
  mexPrintf("Checking contraints.\n") ;
#endif
  for(j=0;j<nd;j++) { /* for all rows */
    if (CJ(1)>CJ(2) || CJ(0)>CJ(2) || CJ(nc-3)>CJ(nc-1) 
	|| CJ(nc-3)>CJ(nc-2)) { 
    contraints_not_met: 
      *E=-100.0 ; return ; }
    for(i=3;i<nc-2;i++) 
      if (CJ(i-1)>CJ(i)) goto contraints_not_met ;
  }
#if DEBUG
  mexPrintf("Contraints met.\n") ;
#endif

  SAFEMALLOC(e,ny,nr) ;  SAFEMALLOC(fwt,ny,nr) ;
  SAFEMALLOC(dereg,1,nc*nd) ;

  /* warp the test image - always Matlab ordering*/  
  imwarpbdd(fwt,ftb,c,nr,nt,nc,nd,ny) ;
#if DEBUG
  mexPrintf("Image warped.\n") ;
#endif

  /* calculate differences and amplitude */
  *E=0.0 ; maxf=0.0 ;
  for(i=0;i<nr;i++) /* columns */
    for(j=0;j<ny;j++) /* rows */
      { tmp=fr[j+ny*i]-fwt[j+ny*i] ; e[j+ny*i]=tmp ; (*E)+=tmp*tmp ; 
        if ((tmp=fabs(fr[j+ny*i]))>maxf) maxf=tmp ;
      }
#if DEBUG
  mexPrintf("Criterion E=%g\n",(double)*E) ;
  mexPrintf("Differences calculated.\n") ;
#endif
  /* zero the elements */
  for(i=0;i<nc;i++) { 
    gr[i]=0.0 ;
    for(j=0;j<nc;j++) H[j*nc+i]=0.0 ;
  }
#if DEBUG
  mexPrintf("Elements zeroed.\n") ;
#endif

  /* now cycle through all the rows */
  for(j=0;j<ny;j++) {
    pary=2+((double)j)/(ny-1)*(nd-3) ;
    for(k=0;k<nr;k++) { /* columns */
      parx=2+((double)k)/(nr-1)*(nc-3) ;
      r=qbinterpd(c,nc,nd,parx,pary) ;
      /* r now contains g(x) - the x coordinate in the test image */
      /* find the first and last influential parameter, counting from 0 ! */
      ll=(int)floor(parx-1.5) ; lh=ll+2 ;
      if (ll<0) ll=0 ; if (lh>=nc) lh=nc-1 ;
      lly=(int)floor(pary-1.5) ; lhy=lly+2 ;
      if (lly<0) lly=0 ; if (lhy>=nd) lhy=nd-1 ;
      
      /* determine influential columns counting from 0 */
      cl=(int)floor(r)-2 ; ch=cl+3 ;
      if (cl<0) cl=0 ; if (ch>=nr) ch=nr-1 ;
      for(der=0.0,l=cl;l<=ch;l++) der-=ftb[j+ny*l]*cbderiv(r-l-1) ;

      /* loop over influential parameters */
      for(i=ll;i<=lh;i++) {
	beta=qbspln(parx-i-1) ;
	for (ii=lly;ii<=lhy;ii++) {
	  betay=qbspln(pary-ii-1) ;
	  gr[ii+i*nd]+=2.0*(dereg[ii+i*nd]=der*beta*betay)*e[j+ny*k] ;
	}
      }

      /* Update Hess matrix, making use of the symmetry */
      for(i=ll;i<=lh;i++) 
	for(iy=lly;iy<=lhy;iy++) 
	  for(ii=i;ii<=lh;ii++) 
	    for(iiy=iy;iiy<=lhy;iiy++) {
	      ind1=i*nd+iy ; ind2=ii*nd+iiy ;
	      if (i==iy && iiy==iy)
		HXY(ind1,ind1)+=2.0*dereg[ind1]*dereg[ind1] ;
	      else {
		tmp=2.0*dereg[ind1]*dereg[ind2] ; 
		HXY(ind1,ind2)+=tmp ; HXY(ind2,ind1)+=tmp ;
	      }
	    }   
    } /* end of the loop across columns */
  }/* end of the loop across rows */
#if DEBUG
  mexPrintf("Derivatives calculated.\n") ;
#endif
  mxFree(e) ; mxFree(fwt) ; mxFree(dereg) ; 
}

