/*
  meximwarpcalcbbtd    The MEX source file

% Given a reference image fr, test image ft, its cubic B-spline
% interpolation coefficients ftb and warp coefficients c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g that is monotonous 
% No penalty function, going out of the permissible space is regarded as
% failure 
% Changed to use quadratic interpolation for g
% Two boundary points of c are not used directly
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns

  Usage: [E,gr,H]=imwarpcalcbbtd(c,fr,ft,ftb)

  Jan Kybic, April 1997

*/

/* #ifdef __GNUC__ 
 #ifndef linux 
 #define linux 
 #endif 
 #endif */

#ifdef linux
#define myCreateMatrix(a,b) mxCreateFull(a,b,REAL)
#define myAlloc(a) mxCalloc(a,1)
#else /* hp version */
#define myCreateMatrix(a,b) mxCreateDoubleMatrix(a,b,mxREAL)
#define myAlloc(a) mxMalloc(a)
#endif

#include <string.h>
#include <math.h>
#include "mex.h"

/* Input Arguments */

#define	C_IN	prhs[0]
#define	FR_IN	prhs[1]
#define	FT_IN	prhs[2]
#define	FTB_IN	prhs[3]

/* Output Arguments */

#define	E_OUT	plhs[0]
#define	GR_OUT	plhs[1]
#define	H_OUT	plhs[2]

#define max(a,b) ((a>b) ? (a) : (b))

void imwarpcalcbbtd(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int ny) ;

#ifdef linux
void mexFunction(
                 int nlhs, Matrix *plhs[],
                 int nrhs, Matrix *prhs[]
		 )
#else
void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
#endif
{ double *c,*fr,*ft,*ftb,*E,*gr,*H ;
  int i,nc,ny,nr,nt ;

  /* Check for proper number of arguments */
  
  if (nrhs != 4) {
    mexErrMsgTxt("IMWARPCALCBBTD requires four input arguments.");
  } else if (nlhs != 3) {
    mexErrMsgTxt("IMWARPCALCBBTD requires three output argument.");
  }
  
  for (i=0;i<nrhs;i++)
    { if (!mxIsNumeric(prhs[i]) || mxIsComplex(prhs[i]) || 
      mxIsSparse(prhs[i])  || !mxIsDouble(prhs[i])) 
         mexErrMsgTxt(
             "IMWARPCALCBBTD needs all input arguments to be real.");
     } 
  
  ny=mxGetM(FR_IN) ; nr=mxGetN(FR_IN) ; nt=mxGetN(FT_IN) ;
  nc=max(mxGetM(C_IN),mxGetN(C_IN)) ;
  
  if (ny!=mxGetM(FT_IN) || ny!=mxGetM(FTB_IN) || nt!=mxGetN(FTB_IN))
    mexErrMsgTxt("IMWARPCALCBBTD - wrong size of arguments.") ;

  fr = mxGetPr(FR_IN); ft=mxGetPr(FT_IN) ; c=mxGetPr(C_IN); 
  ftb=mxGetPr(FTB_IN); 

  E_OUT = myCreateMatrix(1,1); E=mxGetPr(E_OUT) ;
  GR_OUT = myCreateMatrix(1,nc); gr=mxGetPr(GR_OUT) ;
  H_OUT = myCreateMatrix(nc,nc); H=mxGetPr(H_OUT) ;

  /*  %%%%% here the calculation begins %%%%%% */
#if 0
  mexPrintf("imwarpcalcbbtd called nc=%d nr=%d nt=%d ny=%d\n",nc,nr,nt,ny) ;
#endif 
  imwarpcalcbbtd(c,fr,ft,ftb,E,gr,H,nr,nt,nc,ny) ;
}

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt("IMWARPCALCBBTD: could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt("IMWARPCALCBBTD: could not malloc") ;

#define MAKING_MEX

#include "bsplinetools.c"

/* This is the executive function, corresponding to its Matlab
   equivalent */

void imwarpcalcbbtd(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int ny) 
{
  double *e,*fwt,*dereg,*q,tmp,der,r,beta,par,alpha,gamma,aa ;
  double maxf ;
  int i,j,k,ii,ll,lh,cl,ch,l ;

  if (c[0]>c[1] || c[0]>c[2] || c[nc-3]>c[nc-1] || c[nc-2]>c[nc-1]) { 
contraints_not_met: 
    *E=-100.0 ; return ; }
		  

  for(i=3;i<nc-2;i++) 
    if (c[i-1]>c[i]) goto contraints_not_met ;

  SAFEMALLOC(e,ny,nr) ;  SAFEMALLOC(fwt,ny,nr) ;
  SAFEMALLOC(dereg,1,nc) ;

  /* warp the test image - always Matlab ordering*/  
  imwarpbtd(fwt,ftb,c,nr,nt,nc,ny) ;
  /* calculate differences and amplitude */
  *E=0.0 ; maxf=0.0 ;
  for(i=0;i<nr;i++) /* columns */
    for(j=0;j<ny;j++) /* rows */
      { tmp=fr[j+ny*i]-fwt[j+ny*i] ; e[j+ny*i]=tmp ; (*E)+=tmp*tmp ; 
        if ((tmp=fabs(fr[j+ny*i]))>maxf) maxf=tmp ;
      }

  /* zero the elements */
  for(i=0;i<nc;i++) { 
    gr[i]=0.0 ;
    for(j=0;j<nc;j++) H[j*nc+i]=0.0 ;
  }

  /* now cycle through all the rows */
  for(j=0;j<ny;j++) 
    for(k=0;k<nr;k++) { /* columns */
      par=2+((double)k)/(nr-1)*(nc-3) ;
      r=qbinterp(c,nc,par) ;
      /* r now contains g(x) - the x coordinate in the test image */
      /* find the first and last influential parameter, counting from 0 ! */
      ll=(int)floor(par-1.5) ; lh=ll+2 ;
      if (ll<0) ll=0 ; if (lh>=nc) lh=nc-1 ;
      
      /* determine influential columns counting from 0 */
      cl=(int)floor(r)-2 ; ch=cl+3 ;
      if (cl<0) cl=0 ; if (ch>=nr) ch=nr-1 ;
      for(der=0.0,l=cl;l<=ch;l++) der-=ftb[j+ny*l]*cbderiv(r-l-1) ;

      /* loop over influential parameters */
      for(i=ll;i<=lh;i++) {
	beta=qbspln(par-i-1) ;
	gr[i]+=2.0*(dereg[i]=der*beta)*e[j+ny*k] ;
      }

      /* Update Hess matrix, making use of the symmetry */
      for(i=ll;i<=lh;i++) {
	H[i*(nc+1)]+=2.0*dereg[i]*dereg[i] ;
	for(ii=i+1;ii<=lh;ii++) { 
	  tmp=2.0*dereg[i]*dereg[ii] ; 
	  H[i+ii*nc]+=tmp ; H[ii+i*nc]+=tmp ;
	}
      }
    } /* end of the loops across columns and rows */

  mxFree(e) ; mxFree(fwt) ; mxFree(dereg) ; mxFree(q) ;
}

