/*
  meximwarpcalcbdd    The MEX source file

% Given a reference image fr, test image ft, its cubic B-spline
% interpolation coefficients ftb and warp coefficient matrix c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g that is monotonous in both x and y
% Now adds penalty function
% Changed to use quadratic interpolation for g
% Boundary points of c are not used directly
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns. The warping matrix is stored columnwise in a vector, 
% the number of columns given by nc

  Usage: [E,gr,H]=imwarpcalcbdd(c,fr,ft,ftb,nc)

  Jan Kybic, April 1997

*/

/* #ifdef __GNUC__ 
 #ifndef linux 
 #define linux 
 #endif 
 #endif */

#ifdef linux
#define myCreateMatrix(a,b) mxCreateFull(a,b,REAL)
#define myAlloc(a) mxCalloc(a,1)
#else /* hp version */
#define myCreateMatrix(a,b) mxCreateDoubleMatrix(a,b,mxREAL)
#define myAlloc(a) mxMalloc(a)
#endif

#include <string.h>
#include <math.h>
#include "mex.h"

/* Input Arguments */

#define	C_IN	prhs[0]
#define	FR_IN	prhs[1]
#define	FT_IN	prhs[2]
#define	FTB_IN	prhs[3]
#define NC_IN   prhs[4]

/* Output Arguments */

#define	E_OUT	plhs[0]
#define	GR_OUT	plhs[1]
#define	H_OUT	plhs[2]

#define HESSIAN_UPDATE 1
/* incorporate penalty function into the Hessian */
#define USESD 0
/* use the second derivative */

#define max(a,b) ((a>b) ? (a) : (b))

void imwarpcalcbdd(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny) ;

#ifdef linux
void mexFunction(
                 int nlhs, Matrix *plhs[],
                 int nrhs, Matrix *prhs[]
		 )
#else
void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
#endif
{ double *c,*fr,*ft,*ftb,*E,*gr,*H ;
  int i,nc,ny,nr,nt,nd,nct ;

  /* Check for proper number of arguments */
  
  if (nrhs != 5) {
    mexErrMsgTxt("IMWARPCALCBDD requires five input arguments.");
  } else if (nlhs != 3) {
    mexErrMsgTxt("IMWARPCALCBDD requires three output argument.");
  }
  
  for (i=0;i<nrhs;i++)
    { if (!mxIsNumeric(prhs[i]) || mxIsComplex(prhs[i]) || 
      mxIsSparse(prhs[i])  || !mxIsDouble(prhs[i])) 
         mexErrMsgTxt(
             "IWARPCALCBDD needs all input arguments to be real.");
     } 
  
  ny=mxGetM(FR_IN) ; nr=mxGetN(FR_IN) ; nt=mxGetN(FT_IN) ;
  nc=(int)(*(mxGetPr(NC_IN))) ;
  nct=max(mxGetM(C_IN),mxGetN(C_IN)) ;
  nd=nct/nc ;  

  if (ny!=mxGetM(FT_IN) || ny!=mxGetM(FTB_IN) || nt!=mxGetN(FTB_IN))
    mexErrMsgTxt("IMWARPCALCBDD - wrong size of arguments.") ;

  fr = mxGetPr(FR_IN); ft=mxGetPr(FT_IN) ; c=mxGetPr(C_IN); 
  ftb=mxGetPr(FTB_IN); 

  E_OUT = myCreateMatrix(1,1); E=mxGetPr(E_OUT) ;
  GR_OUT = myCreateMatrix(1,nct); gr=mxGetPr(GR_OUT) ;
  H_OUT = myCreateMatrix(nct,nct); H=mxGetPr(H_OUT) ;

  /*  %%%%% here the calculation begins %%%%%% */
#define DEBUG 0
#if DEBUG
  mexPrintf("imwarpcalcbdd called nc=%d nr=%d nt=%d ny=%d\n",nc,nr,nt,ny) ;
#endif 
  imwarpcalcbdd(c,fr,ft,ftb,E,gr,H,nr,nt,nc,nd,ny) ;
#if DEBUG
  mexPrintf("returning E=%g\n",*E) ;
#endif
}

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt("IMWARPCALCBTD: could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt("IMWARPCALCBTD: could not malloc") ;

#define MAKING_MEX

#include "bsplinetools.c"

/* This is the executive function */

void imwarpcalcbdd(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny) 
{
  double *e,*fwt,*dereg,*q,tmp,der,r,beta,alpha,gamma,aa ;
  double maxf,parx,pary,betay ;
  int i,j,k,ii,ll,lh,cl,ch,l,lly,lhy,iy,iiy,ind1,ind2 ;
#if USESD
  double dder,tmp1,tmp2 ;
  double qbsplntabx[3],qbsplntaby[3] ;
#endif

  SAFEMALLOC(e,ny,nr) ;  SAFEMALLOC(fwt,ny,nr) ;
  SAFEMALLOC(dereg,1,nc*nd) ;

  /* warp the test image - always Matlab ordering*/  
  imwarpbdd(fwt,ftb,c,nr,nt,nc,nd,ny) ;
  /* calculate differences and amplitude */
  *E=0.0 ; maxf=0.0 ;
  for(i=0;i<nr;i++) /* columns */
    for(j=0;j<ny;j++) /* rows */
      { tmp=fr[j+ny*i]-fwt[j+ny*i] ; e[j+ny*i]=tmp ; (*E)+=tmp*tmp ; 
        if ((tmp=fabs(fr[j+ny*i]))>maxf) maxf=tmp ;
      }
#if DEBUG
  mexPrintf("Criterion E=%g\n",(double)*E) ;
#endif
  /* zero the elements */
  for(i=0;i<nc;i++) { 
    gr[i]=0.0 ;
    for(j=0;j<nc;j++) H[j*nc+i]=0.0 ;
  }

#define CJ(x) c[j+nd*(x)]
#define CGRJ(x) gr[j+nd*(x)]
#define HXY(y,x) H[(y)+(x)*nc*nd]
#define HCJ(y,x) HXY(j+nd*(y),j+nd*(x))

  /* now cycle through all the rows */
  for(j=0;j<ny;j++) {
    pary=2+((double)j)/(ny-1)*(nd-3) ;
    for(k=0;k<nr;k++) { /* columns */
      parx=2+((double)k)/(nr-1)*(nc-3) ;
      r=qbinterpd(c,nc,nd,parx,pary) ;
      /* r now contains g(x) - the x coordinate in the test image */
      /* find the first and last influential parameter, counting from 0 ! */
      ll=(int)floor(parx-1.5) ; lh=ll+2 ;
      if (ll<0) ll=0 ; if (lh>=nc) lh=nc-1 ;
      lly=(int)floor(pary-1.5) ; lhy=lly+2 ;
      if (lly<0) lly=0 ; if (lhy>=nd) lhy=nd-1 ;
      
      /* determine influential columns counting from 0 */
      cl=(int)floor(r)-2 ; ch=cl+3 ;
      if (cl<0) cl=0 ; if (ch>=nr) ch=nr-1 ;
      for(der=0.0,l=cl;l<=ch;l++) der-=ftb[j+ny*l]*cbderiv(r-l-1) ;
#if USESD
      for(dder=0.0,l=cl;l<=ch;l++) dder-=ftb[j+ny*l]*cbdderiv(r-l-1) ;
#endif
      /* loop over influential parameters */
      for(i=ll;i<=lh;i++) {
#if USESD
	beta=qbsplntabx[i-ll]=qbspln(parx-i-1) ;	
#else
	beta=qbspln(parx-i-1) ;	
#endif
	for (ii=lly;ii<=lhy;ii++) {
#if USESD
	  betay=qbsplntaby[ii-lly]=qbspln(pary-ii-1) ;
#else
	  betay=qbspln(pary-ii-1) ;
#endif
	  gr[ii+i*nd]+=2.0*(dereg[ii+i*nd]=der*beta*betay)*e[j+ny*k] ;
	}
      }

      /* Update Hess matrix, making use of the symmetry */
      for(i=ll;i<=lh;i++) 
	for(iy=lly;iy<=lhy;iy++) {
	  ind1=i*nd+iy ;
#if USESD
	  tmp1=2.0*dder*qbsplntabx[i-ll]*qbsplntaby[iy-lly]*e[j+ny*k] ;
#endif
	  for(ii=i;ii<=lh;ii++) {
#if USESD
	    tmp2=tmp1*qbsplntabx[ii-ll] ;
#endif
	    for(iiy=iy;iiy<=lhy;iiy++) {
	       ind2=ii*nd+iiy ;
	      if (i==iy && iiy==iy)
		HXY(ind1,ind1)+=2.0*dereg[ind1]*dereg[ind1] ;
	      else {
		tmp=2.0*dereg[ind1]*dereg[ind2] ; 
		HXY(ind1,ind2)+=tmp ; HXY(ind2,ind1)+=tmp ;
	      }
#if USESD
	      HXY(ind1,ind2)+=tmp2*qbsplntaby[iiy-lly] ;
#endif
	    } /* for iiy */
	  } /* for ii */ 
	} /* for iy and for i */
    } /* end of the loop across columns */
  }/* end of the loop across rows */

  /* now we shall add the penalty function */
  alpha=(30.0*nc)/nr ; gamma=maxf*nr*ny*1e-6 ; aa=alpha*alpha ;
  SAFEMALLOC(q,1,nc-1) ;

  for (j=0;j<nd;j++) { /* for each row */

    q[0]=gamma*exp(alpha*(CJ(0)-CJ(2))) ;
    q[nc-2]=gamma*exp(alpha*(CJ(nc-3)-CJ(nc-1))) ;
    (*E)+=q[0]+q[nc-2] ;
    for(i=1;i<nc-2;i++) {
      q[i]=gamma*exp(alpha*(CJ(i)-CJ(i+1))) ;
      (*E)+=q[i] ;
    }

    CGRJ(0)+=alpha*q[0] ;   CGRJ(1)+=alpha*q[1] ;  
    CGRJ(2)-=alpha*q[2] ;
    CGRJ(nc-3)+=alpha*q[nc-2] ; CGRJ(nc-2)-=alpha*q[nc-3] ;
    CGRJ(nc-1)-=alpha*q[nc-2] ;

#if HESSIAN_UPDATE
    HCJ(0,0)+=aa*q[0] ; HCJ(0,2)-=aa*q[0] ;
    HCJ(1,1)+=aa*q[1] ; HCJ(1,2)-=aa*q[1] ;
    HCJ(2,0)-=aa*q[0] ; HCJ(2,2)+=aa*q[0] ;
    
    HCJ(nc-3,nc-3)+=aa*q[nc-2] ;
    HCJ(nc-3,nc-1)-=aa*q[nc-2] ;
    HCJ(nc-2,nc-3)-=aa*q[nc-3] ; HCJ(nc-2,nc-2)+=aa*q[nc-3] ;
    HCJ(nc-1,nc-3)-=aa*q[nc-2] ; HCJ(nc-1,nc-1)+=aa*q[nc-2] ;
#endif

    for(i=2;i<nc-2;i++) {
      CGRJ(i)+=alpha*(q[i]-q[i-1]) ;
#if HESSIAN_UPDATE
      HCJ(i,i-1)-=aa*q[i-1] ; HCJ(i,i)+=aa*(q[i-1]+q[i]) ;
      HCJ(i,i+1)-=aa*q[i] ;
#endif
    }
  }

  mxFree(e) ; mxFree(fwt) ; mxFree(dereg) ; mxFree(q) ;
}

