/*
  meximwarpcalcbddh    The MEX source file

% Given a reference image fr, test image ft, its cubic B-spline
% interpolation coefficients ftb and warp coefficient matrix c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g that is monotonous in both x and y
% Now adds penalty function
% Changed to use quadratic interpolation for g
% Boundary points of c are not used directly
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns. The warping matrix is stored columnwise in a vector, 
% the number of columns given by nc
%
% Unlike imwarpcalcbdd, the warp functions are shifted by 0.5 instead of 1 

  Usage: [E,gr,H]=meximwarpcalcbddh(c,fr,ft,ftb,nc)

  Jan Kybic, May 1998

*/

/* #ifdef __GNUC__ 
 #ifndef linux 
 #define linux 
 #endif 
 #endif */

#ifdef linux
#define myCreateMatrix(a,b) mxCreateFull(a,b,REAL)
#define myAlloc(a) mxCalloc(a,1)
#else /* hp version */
#define myCreateMatrix(a,b) mxCreateDoubleMatrix(a,b,mxREAL)
#define myAlloc(a) mxMalloc(a)
#endif

#include <string.h>
#include <math.h>
#include "mex.h"

/* Input Arguments */

#define	C_IN	prhs[0]
#define	FR_IN	prhs[1]
#define	FT_IN	prhs[2]
#define	FTB_IN	prhs[3]
#define NC_IN   prhs[4]

/* Output Arguments */

#define	E_OUT	plhs[0]
#define	GR_OUT	plhs[1]
#define	H_OUT	plhs[2]

#define HESSIAN_UPDATE 1
/* incorporate penalty function into the Hessian */
#define USESD 1
/* use the second derivative */
#define DEGREE 2
/* degree for the splines describing interpolating function */
#define PENDIST (10.0)
/* penalty is considered zero further than this */
#define COMPLEX_PENALTY 1
/* use more complex penalty function calculated for every pixel */

#if DEGREE==2
#define SHIFT (0.5)
#define BSPLN(x) qbspln(x)
#define BINTERPD(c,nc,nd,x,y) qbinterpd(c,nc,nd,x,y)
#define IMWARP(fwt,ftb,c,nr,nt,nc,nd,ny) imwarpbddh(fwt,ftb,c,nr,nt,nc,nd,ny)
#else
#if DEGREE==1
#define SHIFT (0)
#define BSPLN(x) lbspln(x)
#define BINTERPD(c,nc,nd,x,y) lbinterpd(c,nc,nd,x,y)
#define IMWARP(fwt,ftb,c,nr,nt,nc,nd,ny) imwarpldd(fwt,ftb,c,nr,nt,nc,nd,ny)
#else
#error Unsupported degree
#endif
#endif

#define max(a,b) ((a>b) ? (a) : (b))

void imwarpcalcbddh(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny) ;

#ifdef linux
void mexFunction(
                 int nlhs, Matrix *plhs[],
                 int nrhs, Matrix *prhs[]
		 )
#else
void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
#endif
{ double *c,*fr,*ft,*ftb,*E,*gr,*H ;
  int i,nc,ny,nr,nt,nd,nct ;

  /* Check for proper number of arguments */
  
  if (nrhs != 5) {
    mexErrMsgTxt("IMWARPCALCBDDH requires five input arguments.");
  } else if (nlhs != 3) {
    mexErrMsgTxt("IMWARPCALCBDDH requires three output argument.");
  }
  
  for (i=0;i<nrhs;i++)
    { if (!mxIsNumeric(prhs[i]) || mxIsComplex(prhs[i]) || 
      mxIsSparse(prhs[i])  || !mxIsDouble(prhs[i])) 
         mexErrMsgTxt(
             "IWARPCALCBDDH needs all input arguments to be real.");
     } 
  
  ny=mxGetM(FR_IN) ; nr=mxGetN(FR_IN) ; nt=mxGetN(FT_IN) ;
  nc=(int)(*(mxGetPr(NC_IN))) ;
  nct=max(mxGetM(C_IN),mxGetN(C_IN)) ;
  nd=nct/nc ;  

  if (ny!=mxGetM(FT_IN) || ny!=mxGetM(FTB_IN) || nt!=mxGetN(FTB_IN))
    mexErrMsgTxt("IMWARPCALCBDDH - wrong size of arguments.") ;

  fr = mxGetPr(FR_IN); ft=mxGetPr(FT_IN) ; c=mxGetPr(C_IN); 
  ftb=mxGetPr(FTB_IN); 

  E_OUT = myCreateMatrix(1,1); E=mxGetPr(E_OUT) ;
  GR_OUT = myCreateMatrix(1,nct); gr=mxGetPr(GR_OUT) ;
  H_OUT = myCreateMatrix(nct,nct); H=mxGetPr(H_OUT) ;

  /*  %%%%% here the calculation begins %%%%%% */
#define DEBUG 0
#if DEBUG
  mexPrintf("imwarpcalcbddh called nc=%d nr=%d nt=%d ny=%d\n",nc,nr,nt,ny) ;
#endif 
  imwarpcalcbddh(c,fr,ft,ftb,E,gr,H,nr,nt,nc,nd,ny) ;
#if DEBUG
  mexPrintf("returning E=%g\n",*E) ;
#endif
}

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt("IMWARPCALCBDDH: could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt("IMWARPCALCBDDH: could not malloc") ;

#define MAKING_MEX

#include "bsplinetools.c"

/* This is the executive function */

void imwarpcalcbddh(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nr,int nt,int nc,int nd,int ny) 
{
  double *e,*fwt,*dereg,*q,tmp,der,r,beta,alpha,gamma,aa ;
  double maxf,parx,pary,betay ;
  int i,j,k,ii,ll,lh,cl,ch,l,lly,lhy,iy,iiy,ind1,ind2 ;
  double tmp1,tmp2,tmp3,tmp4 ;
#if USESD
  double dder ;
  double qbsplntabx[3],qbsplntaby[3] ;
#endif

#if COMPLEX_PENALTY
  double lastr,qq,lastparx ;
  int lastll ;
#endif

  SAFEMALLOC(e,ny,nr) ;  SAFEMALLOC(fwt,ny,nr) ;
  SAFEMALLOC(dereg,1,nc*nd) ;
  SAFEMALLOC(q,1,nc-1) ;

  /* warp the test image - always Matlab ordering*/  
  IMWARP(fwt,ftb,c,nr,nt,nc,nd,ny) ;
  /* calculate differences and amplitude */
  *E=0.0 ; maxf=0.0 ;
  for(i=0;i<nr;i++) /* columns */
    for(j=0;j<ny;j++) /* rows */
      { tmp=fr[j+ny*i]-fwt[j+ny*i] ; e[j+ny*i]=tmp ; (*E)+=tmp*tmp ; 
        if ((tmp=fabs(fr[j+ny*i]))>maxf) maxf=tmp ;
      }

#if DEBUG
  mexPrintf("Criterion E=%g\n",(double)*E) ;
#endif
  /* zero the elements */
  for(i=0;i<nc*nd;i++) { 
    gr[i]=0.0 ;
    for(j=0;j<nc*nd;j++) H[j*nc+i]=0.0 ;
  }

#define CJ(x) c[j+nd*(x)]
#define CGRIY(x) gr[iy+nd*(x)]
#define CGRJ(x) gr[j+nd*(x)]
#define HXY(y,x) H[(y)+(x)*nc*nd]
#define HCJ(y,x) HXY(j+nd*(y),j+nd*(x))

  /* constants for the penalty function */
  /* was alpha=(30.0*nc)/nr */
#if COMPLEX_PENALTY
  alpha=30 ; gamma=maxf*nr*ny*1e-2 ; aa=alpha*alpha ;
#else
  alpha=(10.0*nc)/nr ; gamma=maxf*nr*ny*1e-6 ; aa=alpha*alpha ;
#endif

  /* now cycle through all the rows */
  for(j=0;j<ny;j++) {
    pary=1+SHIFT+((double)j)/(ny-1)*(nd-1-2*SHIFT) ;
    for(k=0;k<nr;k++) { /* columns */
      parx=1+SHIFT+((double)k)/(nr-1)*(nc-1-2*SHIFT) ;
      r=BINTERPD(c,nc,nd,parx,pary) ;
      /* r now contains g(x) - the x coordinate in the test image */
      /* find the first and last influential parameter, counting from 0 ! */
      ll=(int)floor(parx-1.5) ; lh=ll+2 ;
      if (ll<0) ll=0 ; if (lh>=nc) lh=nc-1 ;
      lly=(int)floor(pary-1.5) ; lhy=lly+2 ;
      if (lly<0) lly=0 ; if (lhy>=nd) lhy=nd-1 ;

      /* determine influential columns counting from 0 */
      cl=(int)floor(r)-2 ; ch=cl+3 ;
      if (cl<0) cl=0 ; if (ch>=nr) ch=nr-1 ;
      for(der=0.0,l=cl;l<=ch;l++) der-=ftb[j+ny*l]*cbderiv(r-l-1) ;
#if USESD
      for(dder=0.0,l=cl;l<=ch;l++) dder-=ftb[j+ny*l]*cbdderiv(r-l-1) ;
#endif
      /* loop over influential parameters */
      for(i=ll;i<=lh;i++) {
#if USESD
	beta=qbsplntabx[i-ll]=BSPLN(parx-i-1) ;	
#else
	beta=BSPLN(parx-i-1) ;	
#endif
	for (ii=lly;ii<=lhy;ii++) {
#if USESD
	  betay=qbsplntaby[ii-lly]=BSPLN(pary-ii-1) ;
#else
	  betay=BSPLN(pary-ii-1) ;
#endif
	  gr[ii+i*nd]+=2.0*(dereg[ii+i*nd]=der*beta*betay)*e[j+ny*k] ;
	}
      }

#if COMPLEX_PENALTY
      if (k>0) {
	qq=alpha*(lastr-r) ;
	if (qq>-PENDIST) {
	  (*E)+= (qq=gamma*exp(qq)) ;
	  qq*=alpha ;
	  for(i=lastll;i<=lh;i++) {
	    tmp=qq*(BSPLN(lastparx-i-1)-BSPLN(parx-i-1)) ;
	    for (iy=lly;iy<=lhy;iy++) {
	      betay=tmp*BSPLN(pary-iy-1) ;
	      CGRIY(i)+=betay ;
#if HESSIAN_UPDATE
	      /* Incorporate penalty into the Hess matrix */
	      tmp3=betay*alpha ;
	      ind1=i*nd+iy ;
	      for(ii=lastll;ii<=lh;ii++) {
		tmp1=tmp3*(BSPLN(lastparx-ii-1)-BSPLN(parx-ii-1)) ;
		for(iiy=lly;iiy<=lhy;iiy++) {
		    ind2=ii*nd+iiy ;
		    tmp2=tmp1*BSPLN(pary-iiy-1) ;
		    HXY(ind1,ind2)+=tmp2 ;
		  } /* for iiy */
	      } /* for iy */
#endif /* HESSIAN_UPDATE */
	    } /* for iy */
	  } /* for i */
	} /* if qq>-PENDIST */
      } /* if k>0 */
      lastll=ll ;
      lastr=r ;
      lastparx=parx ;
#endif

#if USESD
      tmp3=2.0*dder*e[j+ny*k] ;
#endif
      /* Update Hess matrix */
      for(i=ll;i<=lh;i++) {
#if USESD
	tmp4=tmp3*qbsplntabx[i-ll] ;
#endif
	for(iy=lly;iy<=lhy;iy++) {
	  ind1=i*nd+iy ;
#if USESD
	  tmp1=tmp4*qbsplntaby[iy-lly] ;
#endif
	  for(ii=ll;ii<=lh;ii++) {
#if USESD
	    tmp2=tmp1*qbsplntabx[ii-ll] ;
#endif
	    for(iiy=lly;iiy<=lhy;iiy++) {
	       ind2=ii*nd+iiy ;
		tmp=2.0*dereg[ind1]*dereg[ind2] ; 
		HXY(ind1,ind2)+=tmp ; 
#if USESD
		tmp=tmp2*qbsplntaby[iiy-lly] ;
		HXY(ind1,ind2)+=tmp ; 
#endif
	    } /* for iiy */
	  } /* for ii */ 
	} }/* for iy and for i */
    } /* end of the loop across columns */
  }/* end of the loop across rows */

#if COMPLEX_PENALTY==0
  /* now we shall add the penalty function */
  for (j=0;j<nd;j++) { /* for each row */
    for(i=0;i<nc-1;i++) { /* for each column */
      if ((CJ(i)-CJ(i+1))*alpha>-PENDIST) 
      (*E)+= ( q[i]=gamma*exp(alpha*(CJ(i)-CJ(i+1))) ) ;
      else q[i]=0 ;
    } /* for i */

    CGRJ(0)+=alpha*q[0] ;   
    CGRJ(nc-1)-=alpha*q[nc-2] ;

#if HESSIAN_UPDATE
    HCJ(0,0)+=aa*q[0] ; HCJ(0,1)-=aa*q[0] ;
    HCJ(nc-1,nc-2)-=aa*q[nc-2] ; HCJ(nc-1,nc-1)+=aa*q[nc-2] ;
#endif

    for(i=1;i<nc-1;i++) {
      CGRJ(i)+=alpha*(q[i]-q[i-1]) ;
#if HESSIAN_UPDATE
      HCJ(i,i-1)-=aa*q[i-1] ; HCJ(i,i)+=aa*(q[i-1]+q[i]) ;
      HCJ(i,i+1)-=aa*q[i] ;
#endif
    }
  }
#endif /* #if not COMPLEX_PENALTY */

  mxFree(e) ; mxFree(fwt) ; mxFree(dereg) ; mxFree(q) ;
}

