/*
  meximwarpopt    The MEX source file

% Given a test image ft and reference image fr, find warp coefficients c
% so that the total square difference E between the reference image and
% warped test image is minimal
% nc is the number of nodes interpolating the warping function horizontally
% and nd vertically
%
% Uses Levenberg-Marquardt algorithm inspired by Numerical Recipies
% Does not check for constraints but uses penalty function instead
% Now uses linear/quadratic/cubic B-spline interpolation for the warping 
% function g
%
% fr and ft are two dimensional, with the same number of rows. Warping takes
% place across columns
% cin is the starting value for c

  Usage: [c,E]=meximwarpopt(fr,ft,cin,nc,degree)

  Jan Kybic, June 1998

*/

#define myCreateMatrix(a,b) mxCreateDoubleMatrix(a,b,mxREAL)
#define myAlloc(a) mxMalloc(a)

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt(NAME "IMWARPCALCBDDH: could not malloc") ;

#define SAFEFREE(x) mxFree(x)

#include <string.h>
#include <math.h>
#include "mex.h"

/* Input Arguments */

#define	FR_IN	  prhs[0]
#define	FT_IN	  prhs[1]
#define	CIN_IN	  prhs[2]
#define NC_IN     prhs[3]
#define DEGREE_IN  prhs[4]

/* Output Arguments */

#define	C_OUT	plhs[0]
#define	E_OUT	plhs[1]

#define NAME "MEXIMWARPOPT"

#define max(a,b) ((a>b) ? (a) : (b))

void imwarpopt(double *c,double *fr,double *ft,
		   double *E,
                   int nr,int nt,int nc,int nd,int ny,int degree) ;

void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
{ double *c,*fr,*ft,*E,*cin ;
  int i,nc,ny,nr,nt,nd,nct,degree ;

  /* Check for proper number of arguments */
  
  if (nrhs != 5) {
    mexErrMsgTxt(NAME " requires five input arguments.");
  } else if (nlhs != 2) {
    mexErrMsgTxt(NAME " requires two output argument.");
  }
  
  for (i=0;i<nrhs;i++)
    { if (!mxIsNumeric(prhs[i]) || mxIsComplex(prhs[i]) || 
      mxIsSparse(prhs[i])  || !mxIsDouble(prhs[i])) 
         mexErrMsgTxt(
             NAME " needs all input arguments to be real.");
     } 
  
  ny=mxGetM(FR_IN) ; nr=mxGetN(FR_IN) ; nt=mxGetN(FT_IN) ;
  nc=(int)(*(mxGetPr(NC_IN))) ;
  nct=max(mxGetM(CIN_IN),mxGetN(CIN_IN)) ;
  nd=nct/nc ;  
  degree=(int)(*(mxGetPr(DEGREE_IN))) ;

  if (ny!=mxGetM(FT_IN))
    mexErrMsgTxt(NAME " - wrong size of ft.") ;

  fr = mxGetPr(FR_IN); ft=mxGetPr(FT_IN) ; cin=mxGetPr(CIN_IN); 

  E_OUT = myCreateMatrix(1,1); E=mxGetPr(E_OUT) ;
  C_OUT = myCreateMatrix(1,nct); c=mxGetPr(C_OUT) ;

  memcpy(c,cin,nct*sizeof(double)) ;

  /*  %%%%% here the calculation begins %%%%%% */
  imwarpopt(c,fr,ft,E,nr,nt,nc,nd,ny,degree) ;
}


#define MAKING_MEX

#include "bsplinetools.c"
#include "imwarpopt.c"


