/*
  imwarpcalcdc    The C source file, routine called from meximwarpcalcdc

% Given a reference image fr, test image ft, its cubic B-spline
% interpolation coefficients ftb and warp coefficient matrix c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g that is monotonous in both x and y
% Regularization Tikhonov stabilizer.
%
% fr and ft are two dimensional. The 3D warping matrix is stored columnwise 
% in a vector, the number of columns given by nc
%


  Jan Kybic, October 1998

*/

#include <math.h>
#ifndef MAKING_MEX
#include <stdio.h>
#include <stdlib.h>
#define mexErrMsgTxt(x) fprintf(stderr,x) 
#define NAME imwarpcalcdc
#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)malloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEFREE(x) mxFree(x)
#endif /* ifndef MAKING_MEX */

#define MAXSUPPORT 4 /* maximal support length */

static void intmultbspln(double lambda,double *c,int i,int j,int k,int nd,double *E,double *gr) ;
static void tikh(double lambda,double *c,int nc,int nd,double *E,double *gr) ;

/* prototypes needed, description later */


/* c is the 3D matrix of coefficients decribing the warping function
   by using splines of degree `degree', with nc columns
   and nd rows. fr is the reference image with nr columns and ny rows,
   ft is the test image with nt columns and ny rows, while ftb is a
   matrix of cubic B-spline coefficients for interpolating ft. The
   criterion is returned in E, its gradient in gr and Hessian in
   H. All matrices are stored columnwise, i.e. in the Matlab/Fortran
   fashion. lambda is the regularization factor for a Tikhonov stabilizer. */

void imwarpcalcdc(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nrx,int nry,int ntx,int nty, int nc,int nd,float lambda) 
{
  double *e,*fwt,*dereg,*q,tmp,rx,ry,beta,alpha,gamma,aa ;
  double maxf,parx,pary,betay ;
  int i,j,k,ii,ll,lh,cl,ch,rl,rh,l,lly,lhy,iy,iiy,ind1,ind2 ;
  double tmp1,tmp2,tmp3,tmp4,tmp5 ;
  double dderx,ddery,dderxy,derx,dery ;
  double tabx[MAXSUPPORT],taby[MAXSUPPORT] ;
  double dtabx[MAXSUPPORT+1] ;
  double lastr,qq,lastparx ;
  int lastll, support,degree,iq,iiq ;
  double *transc,*transgr ;
  double (*bspln)(), (*binterpd)(), shift ;
  void (*imwarp)() ;


  degree=3 ; /* Hardcoded degree for the warping function so far */
  bspln=cbspln ; imwarp=imwarpcdd ; binterpd=cbinterpd ;
  support=degree+1 ;
  shift=(degree-1)/2.0 ;

  /* allocate temporary memory needed */
  SAFEMALLOC(e,nry,nrx) ;  SAFEMALLOC(fwt,nry,nrx) ;
  SAFEMALLOC(dereg,1,nc*nd*2) ;

  /* warp the test image - always Matlab ordering*/  
  imwarpdcdd(fwt,ftb,c,ntx,nty,nrx,nry,nc,nd) ;
  /* calculate differences and amplitude of the images */
  *E=0.0 ; maxf=0.0 ;
  for(i=0;i<nrx;i++) /* columns */
    for(j=0;j<nry;j++) /* rows */
      { tmp1=fr[j+nry*i] ; tmp=tmp1-fwt[j+nry*i] ; 
        e[j+nry*i]=tmp ; (*E)+=tmp*tmp ; 
        tmp1*=tmp1 ;
        if (tmp1>maxf) maxf=tmp1 ;
      }

  /* zero the elements of the gradient and Hessian */
  for(i=0;i<2*nc*nd;i++) { 
    gr[i]=0.0 ;
    for(j=0;j<2*nc*nd;j++) H[j*nc+i]=0.0 ;
  }

#define HXY(y,x) H[(y)+(x)*nc*nd*2]

  /* now cycle through all the rows */
  for(j=0;j<nry;j++) {
    pary=1+shift+((double)j)/(nry-1)*(nd-1-2*shift) ;
    for(k=0;k<nrx;k++) { /* cycle through columns */
      parx=1+shift+((double)k)/(nrx-1)*(nc-1-2*shift) ;
      rx=binterpd(c,nc,nd,parx,pary) ;
      ry=binterpd(c+nc*nd,nc,nd,parx,pary) ;
      /* rx now contains gx(x) - the x coordinate in the test image */
      /* ry now contains gy(x) - the y coordinate in the test image */
      /* find the first and last influential parameter, counting from 0 ! */
      /* Influential c are then (ll \to lh)\times(lly \to lhy) */
      ll=(int)floor(parx-shift-1) ; lh=ll+support-1 ;
      if (ll<0) ll=0 ; if (lh>=nc) lh=nc-1 ;
      lly=(int)floor(pary-shift-1) ; lhy=lly+support-1 ;
      if (lly<0) lly=0 ; if (lhy>=nd) lhy=nd-1 ;

      /* determine influential columns counting from 0 */
      cl=(int)floor(rx)-2 ; ch=cl+3 ;
      /* The magic constants 2 and 3 are here because the interpolating spline
	 for the image is cubic */
      if (cl<0) cl=0 ; if (ch>=nrx) ch=nrx-1 ;

      /* determine influential rows counting from 0 */
      rl=(int)floor(ry)-2 ; rh=rl+3 ;
      if (rl<0) rl=0 ; if (rh>=nry) rh=nry-1 ;

      /* calculate first and second derivatives of ft at the given coordinates 
         with respect to x and y. Stored as negative. */
      for(derx=0.0,dderx=0.0,l=cl;l<=ch;l++) {
	tmp1=cbderiv(rx-l-1) ; tmp2=cbdderiv(rx-l-1) ;
	for(i=rl;i<=rh;i++) {
	  derx-=tmp1*(tmp3=ftb[i+nry*l]*cbspln(ry-i-1)) ;
	  dderx-=tmp3*tmp2 ;
	}
      }
      for(dery=0.0,ddery=0.0,dderxy=0.0,i=rl;i<=rh;i++) {
	tmp1=cbderiv(ry-i-1) ; tmp2=cbdderiv(ry-i-1) ;
	for(l=cl;l<=ch;l++) {
	  dery-=tmp1*(tmp3=ftb[i+nry*l]*cbspln(rx-l-1)) ;
	  ddery-=tmp3*tmp2 ;
          dderxy-=tmp1*cbderiv(rx-l-1)*ftb[i+nry*l] ;
	}
      }

      /* loop over influential parameters */
      for(i=ll;i<=lh;i++) {
	beta=tabx[i-ll]=bspln(parx-i-1) ; 
	tmp1=beta*derx ;  tmp2=beta*dery ;
	for (ii=lly;ii<=lhy;ii++) {
	  betay=taby[ii-lly]=bspln(pary-ii-1) ;
	  gr[ii+i*nd]+=2.0*(dereg[ii+i*nd]=tmp1*betay)*e[j+nry*k] ;
	  /* with respect to y-components of c */
	  gr[nc*nd+ii+i*nd]+=2.0*(dereg[ii+i*nd+nc*nd]=tmp2*betay)*e[j+nry*k] ;
	}
      }

      /* Update Hess matrix */
      /* iq and iiq distinguish between x/y components of c resp. G resp. H */
      for(iq=0;iq<2;iq++) 
	for(iiq=0;iiq<2;iiq++) {
	  if (iq)
	    { if (iiq) tmp1=ddery ; else tmp1=dderxy ; }  
	  else { if (iiq) tmp1=dderxy ; else tmp1=dderx ; }  
	  tmp2=2.0*e[j+nry*k]*tmp1 ;
	  for(i=ll;i<=lh;i++) {
	    tmp3=tmp2*tabx[i-ll] ;
	    for(iy=lly;iy<=lhy;iy++) {
	      tmp4=tmp3*taby[iy-lly] ;
	      ind1=i*nd+iy+iq*nd*nc ;
	      for(ii=ll;ii<=lh;ii++) {
		tmp5=tmp4*tabx[ii-ll] ;
		for(iiy=lly;iiy<=lhy;iiy++) {
		  ind2=ii*nd+iiy+iiq*nd*nc ;
		  tmp=2.0*dereg[ind1]*dereg[ind2] ; 
		  tmp+=tmp5*taby[iy-lly] ;
		  HXY(ind1,ind2)+=tmp ; 
		} /* for iiy */
	      } /* for ii */ 
	    }/* for iy  */
	  } /* for i */
	} /* for iq and iiq */
   
    } /* end of the loop across columns */
  }/* end of the loop across rows */

  SAFEFREE(e) ; SAFEFREE(fwt) ;  

  /* here we shall add the Tikhonov stabilizer regularization factor */

  lambda*=maxf*nrx*nry/(nc-2)/(nd-2) ; 
  /* normalize lambda according to the number of coefs. */ 

  tikh(lambda,c,nc,nd,E,gr) ; /* x components with respect to x */
  tikh(lambda,c+nc*nd,nc,nd,E,gr+nc*nd) ; /* y components with respect to x */

  SAFEMALLOC(transc,nc,nd*2) ;  /* transposed version of c */
  SAFEMALLOC(transgr,nc,nd*2) ; /* transposed version of gr */
  
  for(i=0;i<nc;i++)
    for(j=0;j<nd;j++) {
      transc[nc*j+i]=c[nd*i+j] ; transc[nc*j+i+nc*nd]=c[nd*i+j+nc*nd] ;
      transgr[nc*j+i]=gr[nd*i+j] ; transgr[nc*j+i+nc*nd]=gr[nd*i+j+nc*nd] ;
    }

  tikh(lambda,transc,nd,nc,E,gr) ; /* x components with respect to y */
  tikh(lambda,transc+nc*nd,nd,nc,E,gr+nc*nd) ; /* y components with respect to y */

  for(i=0;i<nc;i++)
    for(j=0;j<nd;j++) {
      c[nd*i+j]=transc[nc*j+i] ; c[nd*i+j+nc*nd]=transc[nc*j+i+nc*nd] ;
      gr[nd*i+j]=transgr[nc*j+i] ; gr[nd*i+j+nc*nd]=transgr[nc*j+i+nc*nd] ;
    }

  SAFEFREE(transc) ; SAFEFREE(transgr) ;
} /* end of imwarpcalcdc */

static void intmultbspln(double lambda,double *c,int i,int j,int k,int nd,
			 double *E,double *gr) 
/* c is stored in Fortran/Matlab ordering. c[row,column]=c[row+column*nd]
   let f1(x)=c[k-1,i]*B(x+1)+c[k,i]*B(x)+c[k+1,i]*B(x-1)+c[k+2,i]*B(x-2) and
       f2(x)=c[k-1,j]*B(x+1)+c[k,j]*B(x)+c[k+1,j]*B(x-1)+c[k+2,j]*B(x-2), 
   where B() is a centered cubic B-spline.
   It calculates  lambda\int_0^1 f1(x) f2(x) dx and ads it to *E, also updates *gr with 
   the respective derivatives of the regularization criterion with respect to c's.
*/

{ static const double Z[4][4]={{20,129,60,1},{129,1188,933,60},
			       {60,933,1188,129},{1,60,129,20}} ;
  double sum,l ;
  int q,w ;

  l=lambda/5040.0 ;

#if 0 
  mexPrintf("intmultbspln called, i=%d,j=%d,k=%d,nd=%d\n",i,j,k,nd) ;
#endif

  k-- ;
  for (sum=0.0,q=0;q<4;q++) 
    for (w=0;w<4;w++) { 
      sum+=Z[q][w]*c[k+w+j*nd]*c[k+q+i*nd] ;
      gr[k+w+j*nd]+=l*Z[q][w]*c[k+q+i*nd] ;
      gr[k+q+j*nd]+=l*Z[q][w]*c[k+w+i*nd] ;
    }

  *E+=l*sum ;

#if 0
 mexPrintf("intmultbspln returned\n") ;
#endif 
}

static void tikh(double lambda,double *c,int nc,int nd,double *E,double *gr)
{ int i,j ;
  for(j=1;j<nd-2;j++) { /* for all rows except the first and last */
    for(i=1;i<nc-2;i++) {
      intmultbspln(lambda/3.0,c,i-1,i-1,j,nd,E,gr) ;
      intmultbspln(-lambda,c,i-1,i,j,nd,E,gr) ;
      intmultbspln(lambda/3.0,c,i-1,i+2,j,nd,E,gr) ;
      intmultbspln(lambda,c,i,i,j,nd,E,gr) ;
      intmultbspln(-lambda,c,i,i+1,j,nd,E,gr) ;
      intmultbspln(lambda,c,i+1,i+1,j,nd,E,gr) ;
      intmultbspln(-lambda,c,i+1,i+2,j,nd,E,gr) ;
      intmultbspln(lambda/3.0,c,i+2,i+2,j,nd,E,gr) ;
    } /* for i */
  } /* for j */
}
