function [c,E]=imwarpoptdi(fr,ft,nc,nd,c,order) ;
% Given a test image ft and reference image fr, find warp coefficients c
% so that the total square difference E between the reference image and
% warped test image is minimal
% nc is the number of nodes interpolating the warping function horizontally
% and nd vertically. Uses full 2D->2D warping
%
% Uses Levenberg-Marquardt algorithm inspired by Numerical Recipies
% Does not check for constraints but uses regularization function instead
%
% c is the initial estimate of the warping coefficients
% order=1(linear),2(quadratic),3(cubic) (only cubic implemented so far)
%
% Usage: [c,E]=imwarpoptdi(fr,ft,nc,nd,c,order) ;

displaying=1 ; % should progress messages be printed
tracking=0 ;
simplifyH=0 ; % simplify Hessian

Tlambda=0 ; % coefficient for the regularization Tikhonov stabilizer
% Tlambda=1e-5 ;

[ny nr]=size(fr) ;
[nyt nt]=size(ft) ;

if order~=3, error('Unsupported order') ; end ;

if order==2,
  if nc<4 & nd<4, error('nc,nd must be at least 4.') ; end ;
end ;


ftb=imcbanal(ft) ; ftb=imcbanal(ftb')' ;

% [Eold,gr,H]=meximwarpcalcbddh(c,fr,ft,ftb,nc) ;
[Eold,gr,H]=meximwarpcalcdc(c,fr,ft,ftb,nc,Tlambda) ;
if tracking,
  Ehistory=Eold ;
end ;


E=Eold ;
Eold=2*Eold+1 ;
prec=1e-3 ;
delta=1e-10 ;
lambda=0.1 ;
i=1 ; j=0 ;

imh=meximwarpdc(ftb,c,nc,nr,ny) ;


Ein=sum(sum((imh-fr).^2)) ;

while (Eold-E>Eold*prec+delta | j<1), 
  % previously while ((Eold-E)/Eold>prec | j<3),
  if displaying,   
    disp([ 'Iteration ' num2str(i) '  criterion=' num2str(E)]) ; 
  end ;

  E1=E ;
  M=H+eye(nc*nd*2)*lambda ; rc=rcond(M) ;

  if simplifyH,
    % use block diagonalized Hessian
    NN=nc*nd ;
    Ms=zeros(2,2) ; dc=zeros(1,2*NN) ;
    cc=cputime ;
    for ii=1:NN,
      Ms(1,1)=M(ii,ii) ;
      Ms(1,2)=M(ii,ii+NN) ;
      Ms(2,1)=M(ii+NN,ii) ;
      Ms(2,2)=M(ii+NN,ii+NN) ;
      ds=-inv(Ms)*[ gr(ii) ; gr(ii+NN) ] ;
      dc(ii)=ds(1) ; dc(ii+NN)=ds(2) ;
    end ;
    cpdc=c+dc ;
    cc1=cputime ;
    [E,gr1,H1]=meximwarpcalcdc(cpdc,fr,ft,ftb,nc,Tlambda) ;
    cc2=cputime ;
    disp([ 'Inverse took ' num2str(cc1-cc) 's, Warping ' num2str(cc2-cc1) ]) ;
    
    if displaying,
      disp(['Returned E=' num2str(E)]) ;
    end ;	
    i=i+1 ;
    
  else % use full Hessian
    
    if rc>1e-10,
      dc=-gr/M' ;
      cpdc=c+dc ;
      [E,gr1,H1]=meximwarpcalcdc(cpdc,fr,ft,ftb,nc,Tlambda) ;
      
      if displaying,
	disp(['Returned E=' num2str(E)]) ;
      end ;	
      i=i+1 ;
    else
      if displaying,
	disp('Linear equation set close to singular.') ;
      end ;
    end ;
    
  end ; % simplifyH
    
  if E>E1 | ( rc<=1e-10 & simplifyH==0),
    lambda=10*lambda ; E=E1 ;
    if displaying,
      disp(['Increasing lambda and trying again. lambda=' num2str(lambda)]) ;
    end ;  
    j=0 ;
  else
    lambda=lambda/10 ;
    c=cpdc ;
    Eold=E1 ; H=H1 ; gr=gr1 ;
    j=j+1 ;
  end ;

  if tracking,
      Ehistory=[ Ehistory E ] ;
  end ;
end ;

if displaying,
  disp([ 'Iterations ' num2str(i) '  criterion=' num2str(E)]) ;
end ;

disp([ 'Iterations ' num2str(i) '  rel. criterion=' ...
      num2str(E/size(fr,1)/size(fr,2))]) ;
imh=meximwarpdc(ftb,c,nc,nr,ny) ;


Eout=sum(sum((imh-fr).^2)) ;
disp([ 'Criterion ' num2str(Ein) ' -> ' num2str(Eout) ]) ;

if tracking,
  save splnregEhistory Ehistory
end ;
