/*
  meximwarpcalcdc    The MEX source file

% Given a reference image fr, test image ft, its cubic B-spline
% interpolation coefficients ftb and warp coefficient matrix c,
% calculate differences between the reference image and warped test image
% Return E, the sum of the differences squared, and the first (gr) and
% second (H) derivative of E with respect to c
%
% warp coefficients c should define g 2D->2D that is monotonous in both x and y
% The order of approximation for g is 3
%
% fr and ft are two dimensional. The warping 3D-matrix is stored columnwise 
% in a vector, the number of columns given by nc, first all x-coefs than all 
% y-coefs  
%
% lambda is a coefficient for the regularization Tikhonov stabilizer, 

  Usage: [E,gr,H]=meximwarpcalcdc(c,fr,ft,ftb,nc,lambda)

  Jan Kybic, October 1998

*/

#define myCreateMatrix(a,b) mxCreateDoubleMatrix(a,b,mxREAL)
#define myAlloc(a) mxMalloc(a)

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt(NAME ": could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt(NAME "IMWARPCALCBDDH: could not malloc") ;

#define SAFEFREE(x) mxFree(x)

#include <string.h>
#include <math.h>
#include "mex.h"

/* Input Arguments */

#define	C_IN	  prhs[0]
#define	FR_IN	  prhs[1]
#define	FT_IN	  prhs[2]
#define	FTB_IN	  prhs[3]
#define NC_IN     prhs[4]
#define LAMBDA_IN prhs[5]

/* Output Arguments */

#define	E_OUT	plhs[0]
#define	GR_OUT	plhs[1]
#define	H_OUT	plhs[2]

#define NAME "MEXIMWARPCALCDC"

#define max(a,b) ((a>b) ? (a) : (b))

void imwarpcalcdc(double *c,double *fr,double *ft,double *ftb,
		   double *E,double *gr,double *H,
                   int nrx,int nry,int ntx,int nty, int nc,int nd,float lambda) ;

void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
{ double *c,*fr,*ft,*ftb,*E,*gr,*H,lambda ;
  int i,nc,nd,nrx,nry,ntx,nty,nct ;

  /* Check for proper number of arguments */
  
  if (nrhs != 6) {
    mexErrMsgTxt(NAME " requires six input arguments.");
  } else if (nlhs != 3) {
    mexErrMsgTxt(NAME " requires three output argument.");
  }
  
  for (i=0;i<nrhs;i++)
    { if (!mxIsNumeric(prhs[i]) || mxIsComplex(prhs[i]) || 
      mxIsSparse(prhs[i])  || !mxIsDouble(prhs[i])) 
         mexErrMsgTxt(
             NAME " needs all input arguments to be real.");
     } 
  
  nry=mxGetM(FR_IN) ; nrx=mxGetN(FR_IN) ; ntx=mxGetN(FT_IN) ;
  nty=mxGetM(FT_IN) ;
  nc=(int)(*(mxGetPr(NC_IN))) ;
  nct=max(mxGetM(C_IN),mxGetN(C_IN)) ;
  nd=nct/nc/2 ;  
  lambda=(double)(*(mxGetPr(LAMBDA_IN))) ;

  fr = mxGetPr(FR_IN); ft=mxGetPr(FT_IN) ; c=mxGetPr(C_IN); 
  ftb=mxGetPr(FTB_IN); 

  E_OUT = myCreateMatrix(1,1); E=mxGetPr(E_OUT) ;
  GR_OUT = myCreateMatrix(1,nct); gr=mxGetPr(GR_OUT) ;
  H_OUT = myCreateMatrix(nct,nct); H=mxGetPr(H_OUT) ;

  /*  %%%%% here the calculation begins %%%%%% */
  imwarpcalcdc(c,fr,ft,ftb,E,gr,H,nrx,nry,ntx,nty,nc,nd,lambda) ;
}


#define MAKING_MEX

#include "bsplinetools.c"

#include "imwarpcalcdc.c"

/* end of meximwarpcalc */
