function [c,E,nca]=pimwarpoptd(frp,ftp,nc,nd,degree) ;
% Given a test image ft and reference image fr, find warp coefficients c
% so that the total square difference E between the reference image and
% warped test image is minimal
% nc is the number of nodes interpolating the warping function horizontally
% and nd vertically 
%
% degree gives degree of the warping function (actually only 3 is implemented)
%
% Uses Levenberg-Marquardt algorithm inspired by Numerical Recipies
% Does not check for constraints but uses regularization function instead
% Multiresolution pyramid approach
%
% fr and ft are two dimensional, given as pyramids
% In nca it returns the actual final number of columns for c, 
% number of rows must be deduced from the size of c
% 
% Usage: [c,E,nca]=pimwarpopt(frp,ftp,nc,nd,degree) ;


fr=frp{1} ; ft=ftp{1} ;
% Calculate the pyramide level
nl=size(frp,1)-1 ;
% They should have been calculated as follows:
%figure(1) ;
%frp=pyramid(fr,nl) ; 
%figure(2) ;
%ftp=pyramid(ft,nl) ;

imaging=0 ;
dispwfflg=0 ; % show warping function

if degree~=3, 
  error('Unsupported degree for B-spline warping function') ;
end ;

if degree==3,
  nca=4 ; nda=4 ;
elseif degree==2,
  nca=4 ; nda=4 ;
elseif degree==1,
  nca=2 ; nda=2 ;
else
  error('Unsupported degree for B-spline warping function') ;
end ;

cmsf=64 ; % number of colours

[ony onr]=size(fr) ; 
[fullny fullnr]=size(fr) ;

for i=nl:-1:1,
  frc=frp{i} ; ftc=ftp{i} ;
  if i~=nl, 
    oldnr=nr ; oldny=ny ;
  end ;
  ny=size(frc,1) ; nr=size(frc,2) ; nt=size(ftc,2) ; nty=size(frc,1) ;
  disp([ 'Level ' num2str(i) ' rows=' num2str(ny) ' colref=' num2str(nr) ...
	' coltest=' num2str(nt) ' nca=' num2str(nca) ' nda=' num2str(nda)]) ;

if imaging,  
  subplot(221) ; image(cmsf*frc) ; title('reference') ; colorbar ; 
  subplot(222) ; image(cmsf*ftc) ; title('test') ; colorbar ;
  colormap(jet) ;
  pause(0.01) ;  
end ;  

  if i==nl, % The first time -> initialize c
    if degree==3,
      cr=clincf(1,nr) ;
      c=zeros(4,4) ;
      c(:,1)=clincf(cr(1),cr(1)) ;
      c(:,2)=clincf(cr(2),cr(2)) ;
      c(:,3)=clincf(cr(3),cr(3)) ;
      c(:,4)=clincf(cr(4),cr(4)) ;
      dr=clincf(1,ny) ;
      d=zeros(4,4) ;
      d(1,:)=clincf(dr(1),dr(1))' ;
      d(2,:)=clincf(dr(2),dr(2))' ;
      d(3,:)=clincf(dr(3),dr(3))' ;
      d(4,:)=clincf(dr(4),dr(4))' ;
      c=reshape(c,1,nca*nda) ;
      d=reshape(d,1,nca*nda) ;
      c=[c d] ;
    elseif degree==2,
      cr=qlincf(1,nr) ;
      c=zeros(4,4) ;
      c(:,1)=qlincf(cr(1),cr(1)) ;
      c(:,2)=qlincf(cr(2),cr(2)) ;
      c(:,3)=qlincf(cr(3),cr(3)) ;
      c(:,4)=qlincf(cr(4),cr(4)) ;
      c=reshape(c,1,nca*nda) ;
    else
      % linear case
      cr(1)=1; cr(2)=nr ;
      c=[cr(1) cr(1) cr(2) cr(2)] ;
    end ;
    if dispwfflg, figure ; dispwf(c/nr,nca,degree) ; axis off ; end ;
  else
    % image dimension was increased, change c appropriately
    gamma=(nr-1)/(oldnr-1) ;
    delta=1-gamma ;
    c1=c(1:nca*nda) ; c2=c(nca*nda+1:2*nca*nda) ;
    c1=delta+gamma*c1 ;
    gamma=(ny-1)/(oldny-1) ;
    delta=1-gamma ;
    c2=delta+gamma*c2 ;
    c=[c1 c2] ;
  end ;


  
  % Optimize 
  [c,E]=imwarpoptdi(frc,ftc,nca,nda,c,degree) ;
  
    if dispwfflg, figure ; dispwf(c/nr,nca,degree) ; axis off ; end ;
  
  imeb=imcbanal(imcbanal(ftc)')' ;
  imh=meximwarpdc(imeb,c,nca,nr,ny) ;

  
if imaging,  
  subplot(223) ; image(cmsf*imh) ; title('warped') ; colorbar ; grid on ;
  subplot(224) ; image(cmsf*((imh-frc)/2+0.5)) ; title('difference') ; colorbar ; grid on ;
  % disp('Press a key') ; 
  pause(0.01) ;
end ;

  ncaold=nca ; ndaold=nda ;
  % Increase nca, if appropriate
  if degree==3,
    ncn=nca*2-3 ;
  elseif degree==2,
    ncn=nca*2-2 ;
  else
    ncn=nca*2-1 ;
  end;
  
  if ncn<=nc & 4*ncn<=nr,
    nca=ncn ; 
    disp([ 'Expanding parameter space horizontally: nca=' num2str(nca)]) ;
    c1=c(1:ncaold*ndaold) ; c2=c(ncaold*ndaold+1:2*ncaold*ndaold) ;
    c1=reshape(c1,nda,ncaold) ;
    c1=imrexp(c1,degree) ;
    c1=reshape(c1,1,nda*nca) ;
    c2=reshape(c2,nda,ncaold) ;
    c2=imrexp(c2,degree) ;
    c2=reshape(c2,1,nda*nca) ;
    c=[c1 c2] ;
  end ;  

  % Increase nda, if appropriate
  if degree==3,
    ndn=nda*2-3 ;
  elseif degree==2,
    ndn=nda*2-2 ;
  else
    ndn=nda*2-1 ;
  end;

  if ndn<=nd & 4*ndn<=ny,
    nda=ndn ;
    disp([ 'Expanding parameter space vertically: nda=' num2str(nda)]) ;
    c1=c(1:nca*ndaold) ; c2=c(nca*ndaold+1:2*nca*ndaold) ;
    c1=reshape(c1,ndaold,nca) ;
    c1=imrexp(c1',degree)' ;
    c1=reshape(c1,1,nda*nca) ;
    c2=reshape(c2,ndaold,nca) ;
    c2=imrexp(c2',degree)' ;
    c2=reshape(c2,1,nda*nca) ;
    c=[c1 c2] ;
  end ;
  
  if ndaold~=nda | ncaold~=nca,
    % Optimize 
      if dispwfflg, figure ; dispwf(c/nr,nca,degree) ; axis off ; end ;
    [c,E]=imwarpoptdi(frc,ftc,nca,nda,c,degree) ;
    imh=meximwarpdc(imeb,c,nca,nr,ny) ;

    if imaging,
    subplot(223) ; image(cmsf*imh) ; title('warped') ; colorbar ; grid on ;
    subplot(224) ; image(cmsf*((imh-frc)/2+0.5)) ; title('difference') ; colorbar ; grid on ;
    % disp('Press a key') ; 
    pause(0.01) ;
    end ;
  end ;  
  
end ;

if dispwfflg, figure ; dispwf(c/nr,nca,degree) ; axis off ; end ;





 