function [E,g,H,grx,gry,ux,uy]=szcrit(s,fr,ft,m,t)
% Part of the implementation of the Szelinski-Coughlan registration algorithm
% Given two images, reference (fr) and test (ft) as well as the packed node
% positions s, subsampling factor m and basis function type t, it calculates
% the SSD criterion (E), gradient with respect to vector packed g and a
% Hessian approximation H in `block matrix' form, i.e. a series of 2x2
% matrices along the diagonal of the true Hessian arranged in a two-column
% matrix
%
% Jan Kybic, February 1999
%
% Usage: [E,g,H]=szcrit(s,fr,ft,m,t)

[ny,nx]=size(fr) ;    % image size
nsx=floor((nx-2)/m)+1 ; % size of the control grid
nsy=floor((ny-2)/m)+1 ; 
[sx,sy]=szunpack(s,nsx,nsy) ;
Ns=nsx*nsy ; % Number of parameter pairs

[w,lbas,hbas]=szbas(0,0,t) ; % get the support size

% Calculate the displacement field
[ux,uy]=szdispd(sx,sy,nx,ny,m,t) ;

% Warp ft using bilinear interpolation
% fw=imwarp2table(ft,ux,uy,'linear') ;
fw=imwarplinear(ft,ux,uy) ;

% calculate the criterion
e=fw-fr ;
E=sum(sum(e.^2)) ;

% The gradient vector gx1,gy1,gx2,gy2,gx3,gy3,...
g=zeros(2*Ns,1) ;
H=zeros(2*Ns,2) ;

grx=zeros(ny,nx) ;
gry=zeros(ny,nx) ;

fte=[ zeros(1,nx+3) ; zeros(ny,1) ft zeros(ny,2) ; zeros(2,nx+3) ] ;

% k,l is the row, resp. column of the reference image
for k=1:ny,
  for l=1:nx,
    % fetch the coordinates in the test image
    uxc=ux(k,l) ; uyc=uy(k,l) ;
    uxc=max(0,min(nx+1,uxc)) ;
    uyc=max(0,min(ny+1,uyc)) ;
    rx=(l-1.5)/m+1 ; % coordinates in the control grid space
    ry=(k-1.5)/m+1 ;

    % Boundaries of influential control points
    lrx=max(1,ceil(rx-hbas)) ; 
    hrx=min(nsx,floor(rx-lbas)) ;
    lry=max(1,ceil(ry-hbas)) ;
    hry=min(nsy,floor(ry-lbas)) ;
    % if hrx>=lrx & hry>=lry, % ignore points outside  the grid
    % Calculate the gradient of ft at (uxc,uyc)
    entx=floor(uxc) ;  enty=floor(uyc) ; 
    rmx=uxc-entx ; rmy=uyc-enty ; 
    entx=entx+1 ; enty=enty+1 ;
    gx=rmy*(fte(enty+1,entx+1)-fte(enty+1,entx))+...
	(1-rmy)*(fte(enty,entx+1)-fte(enty,entx)) ; 
    grx(k,l)=gx ;
    gy=rmx*(fte(enty+1,entx+1)-fte(enty,entx+1))+...
	(1-rmx)*(fte(enty+1,entx)-fte(enty,entx)) ; 
    gry(k,l)=gy ;
    % For all influential parameters
    for ll=lrx:hrx,
      for kk=lry:hry,
	wij=szbas(rx-ll,ry-kk,t) ;
	ewij=2*e(k,l)*wij ;
	ind=2*((ll-1)*nsy+kk)-1 ;
	% gradient g^u_(ll,kk)
	g(ind)=g(2*((ll-1)*nsy+kk)-1)+gx*ewij ;
	% gradient g^v_(ll,kk)
	g(ind+1)=g(2*((ll-1)*nsy+kk))+gy*ewij ;
	  
	% Calculate the Hessian at the same time
	% only 2x2 block diagonal is computed
	H(ind,1)=H(ind,1)+2*(gx*wij)^2 ;
	H(ind+1,2)=H(ind+1,2)+2*(gy*wij)^2 ;
	H(ind,2)=H(ind,2)+2*gx*gy*wij*wij ;
	H(ind+1,1)=H(ind+1,1)+2*gx*gy*wij*wij ;
      end ; % for ll
    end ;  % for kk
    % else
    %  disp('Out of bounds check active') ;
    % end ; % if hrx>=lrx ...
	
  end ;	% for l 
end ; % for k

