/* ---------------------------------------------------------------------------

Filename		: WaveletTools.c 

Author			: Daniel Sage

Organization	: EPFL, Biomedical Imaging Group

Date	       	: 21/12/1998  

Purpose			: 

	1D signal splitting using a symmetrical filter 
	1D signal merging using a symmetrical filter 

	The wavelet filters should be symmetrical are loaded before running. 
	The boundary conditions can be choosen: Mirror or Peridic.
	
	There are two external functions:
		WaveletSplit_1D()	x[nx] -> y1[nx/2] & y2[nx/2]   
		WaveletMerge_1D()   y1[ny] & y2[ny] -> x[2*ny]

Convention		:

 	v[nv] : lowpass filter, array of coefficients 
 	w[nw] : highpass filter, array of coefficients 

---------------------------------------------------------------------------- */


/* ------------------------------------------------------------------------- */
/* Declaration of extern procedures                                          */
/* ------------------------------------------------------------------------- */
#include <stdlib.h>
#include <string.h>
#include "configs.h"
#include "messagedisplay.h"
#include "wavelettools.h"

/* ------------------------------------------------------------------------- */
/* Declaration of static procedures                                          */
/* ------------------------------------------------------------------------- */
static void WaveletSplit_1DMirror(	
					double x[], long nx, 
					double y1[],double y2[], long *ny,
					double v[], long nv, 
					double w[], long nw);
					
static void WaveletSplit_1DPeriodic(
					double x[], long nx,
					double y1[],double y2[], long *ny,
					double v[], long nv, 
					double w[], long nw);

static void WaveletMerge_1DMirror(	
					double y1[],double y2[],long ny,
					double x[],long *nx,
					double v[],long nv,
					double w[],long nw);
				
static void WaveletMerge_1DPeriodic(	
					double y1[],double y2[],long ny,
					double x[],long *nx,
					double v[],long nv,
					double w[],long nw);


/* ----------------------------------------------------------------------------

	Function:	WaveletSplit_1D
	
	Purpose:	Two channel filtering and down-sampling for 1D real signals. 
				Interface to call the basic routines:
					WaveletSplit_1DMirror() if BoundaryConditions is 1
					WaveletSplit_1DPeriodic() if BoundaryConditions is 2

				lowpass filter:  v[nv]
				highpass filter: w[nw]

				x[nx] -> y1[nx/2] & y2[nx/2]
 

	Author:		Daniel Sage, EPFL, Biomedical Imaging Group
					
---------------------------------------------------------------------------- */
extern int WaveletSplit_1D(	
					double x[], long nx,
					double y1[], double y2[], long *ny,
					double v[], long nv, 
					double w[], long nw,
					short BoundaryConditions)
{

	/* --- Check the parameters --- */
	
	if (x == (double *)NULL) {
		MessageDisplay("ERROR - Pointer of input is null");
		return(ERROR);
	}
	
	if (nx <= 1) {
		MessageDisplay("ERROR - Length of input should be greater than 1");
		return(ERROR);
	}
	
	if ((nx/2L)*2L != nx) {  	/* The input length should be even */
		MessageDisplay("ERROR - Length of input should be even");
		return(ERROR);
	}
	
	if (y1 == (double *)NULL) {
		MessageDisplay("ERROR - Pointer to an output is null");
		return(ERROR);
	}
	
	if (y2 == (double *)NULL) {
		MessageDisplay("ERROR - Pointer to an output is null");
		return(ERROR);
	}
	
	if (ny == (long *)NULL) {
		MessageDisplay("ERROR - Pointer to an output is null");
		return(ERROR);
	}
	
	if (v == (double *)NULL) {
		MessageDisplay("ERROR - Pointer to the lowpass filter is null");
		return(ERROR);
	}
	
	if (nv <= 0 ) {
		MessageDisplay("ERROR - Length of the filter be greater than 0");
		return(ERROR);
	}
	
	if (w == (double *)NULL) {
		MessageDisplay("ERROR - Pointer to the highpass filter is null");
		return(ERROR);
	}
	
	if (nw <= 0 ) {
		MessageDisplay("ERROR - Length of the filter be greater than 0");
		return(ERROR);
	}
			
	/* --- Call the processing function */
	
	switch(BoundaryConditions) {
		case 1:
			WaveletSplit_1DMirror( x, nx,y1, y2, ny, v, nv, w, nw);
			break;
		case 2: 
			WaveletSplit_1DPeriodic( x, nx,y1, y2, ny, v, nv, w, nw);
			break;
		default:	
			MessageDisplay("ERROR - Boundary should be Periodic or Mirror");
			return(ERROR);
	}
		 
	return(!ERROR);
}
					
/* ----------------------------------------------------------------------------

	Function: 	WaveletMerge_1D
	
	Purpose: 	Two channel up-sampling and filtering for 1D real signals. 
				Interface to call the basic routines:
					WaveletSplit_1DMirror() if BoundaryConditions is 1
					WaveletSplit_1DPeriodic() if BoundaryConditions is 2

				lowpass filter:  v[nv]
				highpass filter: w[nw]

				y1[ny] & y2[ny] -> x[2*ny]

	Author:		Daniel Sage, EPFL, Biomedical Imaging Group
	
---------------------------------------------------------------------------- */
extern int WaveletMerge_1D(
				double y1[],double y2[],long ny,
				double x[],long *nx,
				double v[],long nv,
				double w[],long nw,
				short BoundaryConditions)
{

	/* --- Check the parameters --- */
	
	if (y1 == (double *)NULL) {
		MessageDisplay("ERROR - Pointer of input is null");
		return(ERROR);
	}
	
	if (y2 == (double *)NULL) {
		MessageDisplay("ERROR - Pointer of input is null");
		return(ERROR);
	}
	
	if (ny <= 1) {
		MessageDisplay("ERROR - Length of input should be greater than 1");
		return(ERROR);
	}
	
	if (nx == (long *)NULL) {
		MessageDisplay("ERROR - Pointer to an output is null");
		return(ERROR);
	}
	
	if (nx == (long *)NULL) {
		MessageDisplay("ERROR - Pointer to an output is null");
		return(ERROR);
	}
	
	if (v == (double *)NULL) {
		MessageDisplay("ERROR - Pointer to the lowpass filter is null");
		return(ERROR);
	}
	
	if (nv <= 0 ) {
		MessageDisplay("ERROR - Length of the filter be greater than 0");
		return(ERROR);
	}
	
	if (w == (double *)NULL) {
		MessageDisplay("ERROR - Pointer to the highpass filter is null");
		return(ERROR);
	}
	
	if (nw <= 0 ) {
		MessageDisplay("ERROR - Length of the filter be greater than 0");
		return(ERROR);
	}
	
	/* --- Call the processing function */
	
	switch(BoundaryConditions) {
		case 1:		
			WaveletMerge_1DMirror( y1, y2, ny, x, nx, v, nv, w, nw);
			break;
			
		case 2:
			WaveletMerge_1DPeriodic( y1, y2, ny, x, nx, v, nv, w, nw);
			break;
			
		default: 
			MessageDisplay("ERROR - Boundary should be Periodic or Mirror");
			return(ERROR);
		 }
		 
	return(!ERROR);

}

/* ----------------------------------------------------------------------------

	Function:	WaveletSplit_1DMirror
	
	Purpose:	Two channel filtering and down-sampling for 1D
				real signals. The filters are symmetrical and the signal
				is extended using mirror extrapolation.

				lowpass filter:  v[nv]
				highpass filter: w[nw]

				x[nx] -> y1[nx/2] & y2[nx/2]
 
	Author:		Michael Unser, EPFL, Biomedical Imaging Group
					
	History:	MU	Oct 1992	tested
				DS	Dec 1998	renaming, double version  

---------------------------------------------------------------------------- */
static void WaveletSplit_1DMirror(	
					double x[], long nx,
					double y1[], double y2[], long *ny,
					double v[], long nv, 
					double w[], long nw)
{	  
double 	pix;
long 	i, j, k, nyy, n, j1, j2, kn;

	nyy = nx / 2L;
	n   = nyy * 2L;					/* signal length */
	kn  = 2L * n - 2L;				/* Global period */

	if ((nv<=1L)||(nw<=1L))			/* Haar transform */
		for (i=0L;i<nyy;i++) {
			j=2*i;
			y1[i]=(+x[j]+x[j+1])/2.;
			y2[i]=(-x[j]+x[j+1])/2.;
	}
	else
		for (i=0L;i<nyy;i++) {
			j=i*2;
			pix=x[j]*v[0];
			for (k=1;k<nv;k++) {
				j1=j-k;
				j2=j+k;
		
				if (j1<0L) {
					j1=(-j1) % kn;
					if (j1>n-1) j1=kn-j1;
				}
				if (j2>n-1) {
					j2=j2 % kn;
					if (j2>n-1) j2=kn-j2;
				}

				pix=pix+v[k]*(x[j1]+x[j2]);
		}
		y1[i]=pix;

		j=j+1L;
		pix=x[j]*w[0];
		for (k=1L;k<nw;k++) {
			j1=j-k;
			j2=j+k;

			if (j1<0L) {
				j1=(-j1) % kn;
				if (j1>n-1) j1=kn-j1;
			}
			if (j2>n-1) {
				j2=j2 % kn;
				if (j2>n-1L) j2=kn-j2;
			}

			pix=pix+w[k]*(x[j1]+x[j2]);
		}
		y2[i]=pix;
	}
	*ny=nyy;
}

/* ----------------------------------------------------------------------------

	Function: 	WaveletSplit_1DPeriodic
	
	Purpose: 	Two channel filtering and down-sampling for 1D
				real signals. The filters are symmetrical and the signal
				is extended using periodic boundary conditions.
				
				lowpass filter:  v[nv]
				highpass filter: w[nw]

				x[nx] -> y1[nx/2] & y2[nx/2]

	Author:		Michael Unser, EPFL, Biomedical Imaging Group
					
	History:	MU	Oct 1992	tested
				DS	Dec 1998	renaming, double version  

---------------------------------------------------------------------------- */
static void WaveletSplit_1DPeriodic(
					double x[], long nx,
					double y1[], double y2[], long *ny,
					double v[], long nv, 
					double w[], long nw)
{	  
double	pix;
long	i, j, k, nyy, n, j1, j2, kn, knn;

	nyy=nx/2;
	n=nyy*2;						/* signal length */
	kn=n;						    /* Global period */
	knn=20*n;

	if ((nv<=1)||(nw<=1))			/* Haar transform */
	for (i=0;i<nyy;i++) {
		j=2*i;
		y1[i]=(+x[j]+x[j+1])/2.;
		y2[i]=(-x[j]+x[j+1])/2.;
	}
	else
	for (i=0;i<nyy;i++) {
		j=i*2;
		pix=x[j]*v[0];
		for (k=1;k<nv;k++) {
			j1=j-k;
			j2=j+k;
			if (j1<0) j1=(knn+j1) % kn;
			if (j2>n-1) j2=j2 % kn;
			pix=pix+v[k]*(x[j1]+x[j2]);
		}
		y1[i]=pix;

		j=j+1;
		pix=x[j]*w[0];
		for (k=1;k<nw;k++) {
			j1=j-k;
			j2=j+k;
			if (j1<0) j1=(knn+j1) % kn;
			if (j2>n-1) j2=j2 % kn;
			pix=pix+w[k]*(x[j1]+x[j2]);
		}
		y2[i]=pix;
	}
	*ny=nyy;
}


/* ----------------------------------------------------------------------------

	Function: 	WaveletMerge_1DMirror
	
	Purpose: 	Two channel up-sampling and filtering for 1D
				real signals. The filters are symmetrical and the
				boundary conditions are mirror.

				lowpass filter:  v[nv]
				highpass filter: w[nw]

				y1[ny] & y2[ny] -> x[2*ny]

 	Author:		Michael Unser, EPFL, Biomedical Imaging Group
					
	History:	MU	Oct 1992	tested
                MU	Apr 1997	corrected boundary conditions)
				DS	Dec 1998	renaming, double version  

---------------------------------------------------------------------------- */
static void WaveletMerge_1DMirror(
				double y1[],double y2[],long ny,
				double x[],long *nx,
				double v[],long nv,
				double w[],long nw)
{	  
double	pix1, pix2;
long	i, j, k, kk, i1, i2, k01, k02, n, kn;

	*nx=ny*2;
	k01=(nv/2)*2-1;
	k02=(nw/2)*2-1;
	n=ny;
	kn=2*n-1;

	if ((nv<=1)||(nw<=1))				/* Haar transform */
		for (i=0;i<ny;i++) {
			j=2*i;
			x[j]=y1[i]-y2[i];
			x[j+1]=y1[i]+y2[i];
		}
	else
	for (i=0;i<ny;i++)	{
	      j=2*i;
	      pix1=v[0]*y1[i];
	      for (k=2;k<nv;k=k+2) {
	        i1=i-(k/2);
	        i2=i+(k/2);
			if (i1<0) {					/* standard boundary */
			  i1=(-i1) % kn;
			  if (i1>n-1) i1=kn-i1;
			}
			if (i2>n-1) {				/* non-standard boundary */
			  i2=(i2) % kn;
			  if (i2>n-1) i2=kn-i2;
			}
			/* NOTE : This is the correct boundary condition assuming that 
			the original image has an even size (2n) and the decimation 
			was performed according to the sequence 1,3,5,,2n-1.	*/
	        pix1=pix1+v[k]*(y1[i1]+y1[i2]);
	      }

		pix2=0.;
		for (k=-k02;k<nw;k=k+2) {
			kk=abs(k);
			i1=i+(k-1)/2;
			if (i1<0) {					/* non-standard boundary */
				i1=(-i1-1) % kn;
				if (i1>n-1) i1=kn-1-i1;
			}
			if (i1>n-1) {				/* standard boundary */
				i1=i1 % kn;
				if (i1>n-1) i1=kn-1-i1;
			}
			/* NOTE : This is the correct boundary condition assuming that 
			the original image has an even size (2n) and the decimation 
			was performed according to the sequence 2,4,6,,2n.	*/
			pix2=pix2+w[kk]*y2[i1];
		}

		x[j]=(pix1+pix2);

		j=j+1;
		pix1=0.;
		for (k=-k01;k<nv;k=k+2) {
			kk=abs(k);
			i1=i+(k+1)/2;
			if (i1<0) {					/* standard boundary */
				i1=(-i1) % kn;
				if (i1>n-1) i1=kn-i1;
			}
			if (i1>n-1) {				/* non-standard boundary */
				i1=(i1) % kn;
				if (i1>n-1) i1=kn-i1;
			}
			pix1=pix1+v[kk]*y1[i1];
		}
		pix2=w[0]*y2[i];
		for (k=2;k<nw;k=k+2) {
			i1=i-(k/2);
			i2=i+(k/2);
			if (i1<0) {					/* non-standard boundary */
				i1=(-i1-1) % kn;
				if (i1>n-1) i1=kn-1-i1;
			}
			if (i2>n-1) {				/* standard boundary */
				i2=i2 % kn;
				if (i2>n-1) i2=kn-1-i2;
			}
			pix2=pix2+w[k]*(y2[i1]+y2[i2]);
		}
		x[j]=(pix1+pix2);
	}
}

/* ----------------------------------------------------------------------------

	Function:	WaveletMerge_1DPeriodic
	
	Purpose:	Two channel up-sampling and filtering for 1D
				real signals. The filters are symmetrical and the
				boundary conditions are periodic.
	
				lowpass filter:  v[nv]
				highpass filter: w[nw]

				y1[ny] & y2[ny] -> x[2*ny]
  
	Author:		Michael Unser, EPFL, Biomedical Imaging Group
					
	History:	MU	Oct 1992	tested
				DS	Dec 1998	renaming, double version  


---------------------------------------------------------------------------- */
static void WaveletMerge_1DPeriodic(
				double y1[],double y2[], long ny, 
				double x[], long *nx, 
				double v[],long nv,
				double w[],long nw)
{	  
double 	pix1, pix2;
long	 i, j, k, kk, i1, i2, k01, k02, n, kn, knn;

	*nx=ny*2;
	k01=(nv/2)*2-1;
	k02=(nw/2)*2-1;
	n=ny;
	kn=n;
	knn=20*n;

	if ((nv<=1)||(nw<=1))				/* Haar transform */
		for (i=0;i<ny;i++) {
			j=2*i;
			x[j]=y1[i]-y2[i];
			x[j+1]=y1[i]+y2[i];
		}
	else
	for (i=0;i<ny;i++)	{
		j=2*i;
		pix1=v[0]*y1[i];
		for (k=2;k<nv;k=k+2) {
			i1=i-(k/2);
			i2=i+(k/2);

			if (i1<0) i1=(knn+i1) % kn;
			if (i2>n-1) i2=i2 % kn;

			pix1=pix1+v[k]*(y1[i1]+y1[i2]);
		}

		pix2=0.;
		for (k=-k02;k<nw;k=k+2) {
			kk=abs(k);
			i1=i+(k-1)/2;
			if (i1<0) i1=(knn+i1) % kn;
			if (i1>n-1) i1=i1 % kn;
			pix2=pix2+w[kk]*y2[i1];
		}
		x[j]=(pix1+pix2);

		j=j+1;
		pix1=0.;
		for (k=-k01;k<nv;k=k+2) {
			kk=abs(k);
			i1=i+(k+1)/2;
			if (i1<0) i1=(knn+i1) % kn;
			if (i1>n-1) i1=i1 % kn;
			pix1=pix1+v[kk]*y1[i1];
		}
		pix2=w[0]*y2[i];
		for (k=2;k<nw;k=k+2) {
			i1=i-(k/2);
			i2=i+(k/2);
			if (i1<0) i1=(knn+i1) % kn;
			if (i2>n-1) i2=i2 % kn;
			pix2=pix2+w[k]*(y2[i1]+y2[i2]);
		}
		x[j]=(pix1+pix2);
	}
}



